"""Unit tests for input structures."""

import pickle
from itertools import chain
from typing import Any, ClassVar

import pytest

from py21cmfast import (
    AstroOptions,
    AstroParams,
    CosmoParams,
    InputParameters,
    MatterOptions,
    SimulationOptions,
    config,
)
from py21cmfast import _templates as tmpl
from py21cmfast.input_serialization import (
    deserialize_inputs,
    prepare_inputs_for_serialization,
)
from py21cmfast.wrapper.inputs import CosmoTables, Table1D

_TEMPLATES = tmpl.list_templates()
_ALL_ALIASES = list(chain.from_iterable(t["aliases"] for t in _TEMPLATES))


class TestInputStructSubclasses:
    """Tests of the InputStruct object and its subclasses."""

    def setup_class(self):
        """Set up basic objects to have fun with."""
        self.cosmo = CosmoParams(SIGMA_8=0.8)
        self.cosmo2 = CosmoParams(SIGMA_8=0.9)

    def test_hash(self):
        """Test that hashes are unequal beween different objects."""
        assert hash(self.cosmo) != hash(self.cosmo2)

    def test_equality(self):
        """Ensure that the python dict has all fields."""
        assert self.cosmo == self.cosmo
        assert self.cosmo != self.cosmo2

    def test_identity(self):
        """Check that the 'is' operator works as expected."""
        assert self.cosmo is self.cosmo
        assert self.cosmo is not self.cosmo2

        c = CosmoParams(SIGMA_8=0.8)
        assert self.cosmo is not c

    def test_bad_construction(self):
        """Test basic errors of construction via __init__."""
        with pytest.raises(TypeError):
            CosmoParams(self.cosmo, self.cosmo)

        with pytest.raises(TypeError):
            CosmoParams(SimulationOptions())

        with pytest.raises(TypeError):
            CosmoParams(1)

        with pytest.raises(TypeError):
            CosmoParams.new(None, SIGMA_8=0.8, bad_param=1)

    def test_constructed_from_itself(self):
        """Test that constructing an object from itself returns a new equivalent one."""
        c3 = CosmoParams.new(self.cosmo)

        assert self.cosmo == c3
        assert self.cosmo is not c3

    def test_altered_construction(self):
        """Test that altering params upon construction alters the output object."""
        c3 = CosmoParams.new(self.cosmo, SIGMA_8=0.7)

        assert self.cosmo != c3
        assert c3.SIGMA_8 == 0.7

    def test_clone(self):
        """Test that the .clone() method returns a distinct but equivalent clone."""
        u = SimulationOptions()
        v = u.clone()
        assert u == v
        assert u is not v

    def test_repr(self):
        """Test that repr works."""
        assert "SIGMA_8=0.8" in repr(self.cosmo)

    def test_pickle(self):
        """Test that pickling works (important for paralellization)."""
        # Make sure we can pickle/unpickle it.
        c4 = pickle.dumps(self.cosmo)
        c4 = pickle.loads(c4)
        assert self.cosmo == c4

        # Make sure the c data gets loaded fine.
        assert c4.cstruct.hlittle == self.cosmo.cstruct.hlittle

    def test_asdict(self):
        """Test the asdict() method works."""
        c5 = CosmoParams.new(self.cosmo)
        assert c5 == self.cosmo
        assert c5.asdict() == self.cosmo.asdict()
        assert c5.cdict == self.cosmo.cdict
        assert (
            c5.cdict != c5.asdict()
        )  # not the same because the former doesn't include dynamic parameters.
        assert c5 == self.cosmo


class TestCosmoTables:
    """Tests of CosmoTables."""

    def setup_class(self):
        """Set up basic objects to have fun with."""
        self.size = 3
        self.x_values = [1, 2, 3]
        self.y_values = [4, 5, 6]
        self.transfer_density = (
            Table1D(
                size=self.size,
                x_values=self.x_values,
                y_values=self.y_values,
            ),
        )
        self.transfer_vcb = Table1D(
            size=self.size,
            x_values=self.x_values,
            y_values=self.y_values,
        )
        self.cosmo_tables = CosmoTables(
            transfer_density=self.transfer_density, transfer_vcb=self.transfer_vcb
        )

    def test_constructed_from_itself(self):
        """Test that constructing an object from itself returns a new equivalent one."""
        cosmo_tables2 = CosmoTables.new(self.cosmo_tables)
        cosmo_tables3 = CosmoTables.new(
            {
                "transfer_density": self.cosmo_tables.transfer_density,
                "transfer_vcb": self.cosmo_tables.transfer_vcb,
            }
        )

        assert self.cosmo_tables == cosmo_tables2
        assert self.cosmo_tables == cosmo_tables3
        assert self.cosmo_tables is not cosmo_tables2
        assert self.cosmo_tables is not cosmo_tables3

    def test_none_input_to_new(self):
        """Test None input to a new CosmoTables."""
        cosmo_tables2 = CosmoTables.new()
        cosmo_tables3 = CosmoTables.new()

        assert self.cosmo_tables != cosmo_tables2
        assert cosmo_tables2 == cosmo_tables3
        assert cosmo_tables2 is not cosmo_tables3

    def test_bad_input(self):
        """Test bad inputs."""
        with pytest.raises(ValueError, match="Cannot instantiate"):
            CosmoParams.new("")


class TestCosmoParams:
    """Tests of CosmoParams."""

    sigma_8 = 1.0
    A_s = 3.0e-9

    def test_defaults(self):
        """Test defaults."""
        cosmo_params = CosmoParams()
        assert cosmo_params.SIGMA_8 == cosmo_params._DEFAULT_SIGMA_8
        assert cosmo_params.A_s == cosmo_params._DEFAULT_A_s

    def test_sigma8(self):
        """Test defaults with sigma8."""
        cosmo_params = CosmoParams(SIGMA_8=self.sigma_8)
        assert self.sigma_8 == cosmo_params.SIGMA_8
        assert cosmo_params.A_s != cosmo_params._DEFAULT_A_s

    def test_A_s(self):
        """Test defaults with A_s."""
        cosmo_params = CosmoParams(A_s=self.A_s)
        assert cosmo_params.SIGMA_8 != cosmo_params._DEFAULT_SIGMA_8
        assert cosmo_params.A_s == self.A_s

    def test_bad_input(self):
        """Test bad inputs."""
        with pytest.raises(ValueError, match="Cannot set both SIGMA_8 and A_s!"):
            CosmoParams(SIGMA_8=self.sigma_8, A_s=self.A_s)


class TestAstroOptions:
    """Tests of AstroOptions."""

    # Testing all the AstroOptions dependencies, including emitted warnings
    def test_bad_inputs(self):
        """Test possible exceptions when creating the object."""
        with pytest.raises(
            ValueError,
            match="You have set USE_MINI_HALOS to True but INHOMO_RECO is False!",
        ):
            AstroOptions(USE_MINI_HALOS=True, USE_TS_FLUCT=True, INHOMO_RECO=False)

        with pytest.raises(
            ValueError,
            match="You have set USE_MINI_HALOS to True but USE_TS_FLUCT is False!",
        ):
            AstroOptions(USE_MINI_HALOS=True, INHOMO_RECO=True, USE_TS_FLUCT=False)

        msg = r"USE_MINI_HALOS is not compatible with the redshift-based"
        with pytest.raises(ValueError, match=msg):
            AstroOptions(
                PHOTON_CONS_TYPE="z-photoncons",
                USE_MINI_HALOS=True,
                INHOMO_RECO=True,
                USE_TS_FLUCT=True,
            )

        with pytest.raises(
            ValueError, match="USE_EXP_FILTER is True but CELL_RECOMB is False"
        ):
            AstroOptions(USE_EXP_FILTER=True, CELL_RECOMB=False)

        with pytest.raises(
            ValueError,
            match="USE_EXP_FILTER can only be used with a real-space tophat HII_FILTER==0",
        ):
            AstroOptions(USE_EXP_FILTER=True, HII_FILTER="sharp-k")


class TestSimulationOptions:
    """Test the SimulationOptions class."""

    def test_non_cubic(self):
        """Test setting non_cubic_factor."""
        up = SimulationOptions()
        up_non_cubic = SimulationOptions(NON_CUBIC_FACTOR=1.5)

        assert up_non_cubic.tot_fft_num_pixels == 1.5 * up.tot_fft_num_pixels
        assert up_non_cubic.HII_tot_num_pixels == up.HII_tot_num_pixels * 1.5
        assert up.cell_size / up.cell_size_hires == up.DIM / up.HII_DIM

    def test_bad_non_cubic_value(self):
        """Test exceptions when setting non_cubic_factor with bad values."""
        with pytest.raises(
            ValueError,
            match="NON_CUBIC_FACTOR \\* DIM and NON_CUBIC_FACTOR \\* HII_DIM must be integers",
        ):
            SimulationOptions(NON_CUBIC_FACTOR=1.1047642)

    def check_attributes_dim(self, s, checks):
        """Check for attributes relating to DIM.

        This is used in several following tests.
        """
        assert checks[0] == s.HIRES_TO_LOWRES_FACTOR
        assert checks[1] == s._HIRES_TO_LOWRES_FACTOR
        assert checks[2] == s.DIM
        assert checks[3] == s._DIM

    @pytest.mark.parametrize(
        ("options", "expected"),
        [
            pytest.param({}, [3, None, 300, None], id="default"),
            pytest.param({"HIRES_TO_LOWRES_FACTOR": 2}, (2, 2, 200, None), id="direct"),
            pytest.param({"DIM": 200}, (2, None, 200, 200), id="explicit"),
        ],
    )
    def test_dim_setting_direct(self, options: dict, expected: tuple[float | None]):
        """Tests of setting DIM vs. HIRES_TO_LOWRES_FACTOR."""
        # By default, uses ratio of 3:
        s = SimulationOptions(HII_DIM=100, **options)
        self.check_attributes_dim(s, expected)

    def test_dim_setting_exceptions(self):
        """Test that exceptions are raised when bad options are passed for DIM."""
        # Exception: explicitly setting both
        with pytest.raises(
            ValueError, match="Cannot set both DIM and HIRES_TO_LOWRES_FACTOR"
        ):
            SimulationOptions(HII_DIM=100, DIM=200, HIRES_TO_LOWRES_FACTOR=2)

        # Exception: Evolve from ratio --> explicit
        with pytest.raises(
            ValueError, match="Cannot set both DIM and HIRES_TO_LOWRES_FACTOR"
        ):
            SimulationOptions(HII_DIM=100, HIRES_TO_LOWRES_FACTOR=2).clone(DIM=200)

        # Exception: Evolve from explicit -> ratio
        with pytest.raises(
            ValueError, match="Cannot set both DIM and HIRES_TO_LOWRES_FACTOR"
        ):
            SimulationOptions(HII_DIM=100, DIM=200).clone(HIRES_TO_LOWRES_FACTOR=3)

    @pytest.mark.parametrize(
        ("direct", "evolved", "expected"),
        [
            pytest.param(
                {}, {"HIRES_TO_LOWRES_FACTOR": 4}, [4, 4, 400, None], id="unset->ratio"
            ),
            pytest.param({}, {"DIM": 300}, [3, None, 300, 300], id="unset->explicit"),
            pytest.param({}, {}, [3, None, 300, None], id="unset->unset"),
            pytest.param(
                {"DIM": 200}, {"DIM": 300}, [3, None, 300, 300], id="explicit->explicit"
            ),
            pytest.param(
                {"HIRES_TO_LOWRES_FACTOR": 2},
                {"HIRES_TO_LOWRES_FACTOR": 3},
                [3, 3, 300, None],
                id="ratio->ratio",
            ),
        ],
    )
    def test_dim_setting_evolve(self, direct, evolved, expected):
        """Test for correct final output after creation then *evolution*."""
        # Evolve from unset -> ratio
        s = SimulationOptions(HII_DIM=100, **direct).clone(**evolved)
        self.check_attributes_dim(s, expected)

    @pytest.mark.parametrize(
        ("direct", "evolved", "expected"),
        [
            pytest.param(
                {}, {"HIRES_TO_LOWRES_FACTOR": 4}, [4, 4, 400, None], id="unset->ratio"
            ),
            pytest.param({}, {"DIM": 300}, [3, None, 300, 300], id="unset->explicit"),
            pytest.param({}, {}, [3, None, 300, None], id="unset->unset"),
            pytest.param(
                {"DIM": 200}, {"DIM": 300}, [3, None, 300, 300], id="explicit->explicit"
            ),
            pytest.param(
                {"HIRES_TO_LOWRES_FACTOR": 2},
                {"HIRES_TO_LOWRES_FACTOR": 3},
                [3, 3, 300, None],
                id="ratio->ratio",
            ),
        ],
    )
    @pytest.mark.parametrize("mode", ["full", "minimal"])
    def test_dim_setting_serialization(
        self,
        direct: dict[str, Any],
        evolved: dict[str, Any],
        expected: tuple[float | None],
        mode: str,
    ):
        """Tests of setting DIM via explicit or ratios, using deserialization.

        This tests for what happens when you write either a TOML file
        or an HDF5 file, then read it in and want to evolve it.
        """
        # To/From TOML
        s = SimulationOptions(HII_DIM=100, **direct)
        inputs = InputParameters(random_seed=1, simulation_options=s)
        dct = prepare_inputs_for_serialization(inputs, mode=mode)
        new = InputParameters(random_seed=1, **deserialize_inputs(dct))

        assert new == inputs

        # Now evolve
        new = new.evolve_input_structs(**evolved)
        self.check_attributes_dim(new.simulation_options, expected)


class TestMatterOptions:
    """Tests of the MatterOptions class."""

    def test_bad_inputs(self):
        """Test that exceptions are raised for bad inputs."""
        msg = r"SOURCE_MODEL settings using the halo sampler require the use of HMF interpolation tables."
        with pytest.raises(ValueError, match=msg):
            MatterOptions(
                SOURCE_MODEL="CHMF-SAMPLER",
                USE_INTERPOLATION_TABLES="sigma-interpolation",
            )

        msg = r"Can only use 'CLASS' power spectrum with relative velocities"
        with pytest.raises(ValueError, match=msg):
            MatterOptions(USE_RELATIVE_VELOCITIES=True, POWER_SPECTRUM="EH")

        msg = r"The conditional mass functions requied for the discrete halo field"
        with pytest.raises(NotImplementedError, match=msg):
            MatterOptions(SOURCE_MODEL="CHMF-SAMPLER", HMF="WATSON")


class TestInputParameters:
    """Tests of the InputParameters class."""

    EXCEPTION_CASES: ClassVar = [
        (
            ValueError,
            "SOURCE_MODEL == 'CONST-ION-EFF' is not compatible with USE_MINI_HALOS=True",
            {
                "matter_options": MatterOptions(SOURCE_MODEL="CONST-ION-EFF"),
                "astro_options": AstroOptions(
                    USE_MINI_HALOS=True,
                    INHOMO_RECO=True,
                    USE_TS_FLUCT=True,
                    USE_EXP_FILTER=False,
                    USE_UPPER_STELLAR_TURNOVER=False,
                ),
            },
        ),
        (
            ValueError,
            "R_BUBBLE_MAX is larger than BOX_LEN",
            {
                "astro_params": AstroParams(R_BUBBLE_MAX=100),
                "simulation_options": SimulationOptions(BOX_LEN=50),
            },
        ),
        (
            ValueError,
            "Your R_BUBBLE_MAX is > BOX_LEN/3",
            {
                "astro_params": AstroParams(R_BUBBLE_MAX=20),
                "simulation_options": SimulationOptions(BOX_LEN=50),
                "astro_options": AstroOptions(
                    USE_EXP_FILTER=False, HII_FILTER="sharp-k"
                ),
            },
        ),
        (
            ValueError,
            "is not compatible with the redshift-based",
            {
                "matter_options": MatterOptions(SOURCE_MODEL="CHMF-SAMPLER"),
                "astro_options": AstroOptions(PHOTON_CONS_TYPE="z-photoncons"),
            },
        ),
        (
            ValueError,
            "USE_EXP_FILTER is not compatible with SOURCE_MODEL == E-INTEGRAL",
            {
                "matter_options": MatterOptions(
                    SOURCE_MODEL="E-INTEGRAL",
                ),
                "astro_options": AstroOptions(
                    USE_EXP_FILTER=True, USE_UPPER_STELLAR_TURNOVER=False
                ),
            },
        ),
        (
            NotImplementedError,
            "USE_UPPER_STELLAR_TURNOVER is not yet implemented for SOURCE_MODEL",
            {
                "matter_options": MatterOptions(
                    SOURCE_MODEL="L-INTEGRAL",
                ),
                "astro_options": AstroOptions(
                    USE_UPPER_STELLAR_TURNOVER=True, USE_EXP_FILTER=False
                ),
            },
        ),
    ]

    WARNINGS_CASES: ClassVar = [
        (
            "You are setting M_TURN > 8 when USE_MINI_HALOS=True.",
            {
                "astro_params": AstroParams(M_TURN=10),
                "astro_options": AstroOptions(
                    USE_MINI_HALOS=True, USE_TS_FLUCT=True, INHOMO_RECO=True
                ),
            },
        ),
        (
            "Resolution is likely too low for accurate evolved density fields",
            {"simulation_options": SimulationOptions(BOX_LEN=50, DIM=20)},
        ),
        (
            r"This is non\-standard \(but allowed\), and usually occurs upon manual update of INHOMO_RECO",
            {
                "astro_params": AstroParams(R_BUBBLE_MAX=10),
                "simulation_options": SimulationOptions(BOX_LEN=50),
                "astro_options": AstroOptions(INHOMO_RECO=True),
            },
        ),
        (
            "USE_MINI_HALOS needs USE_RELATIVE_VELOCITIES",
            {
                "matter_options": MatterOptions(USE_RELATIVE_VELOCITIES=False),
                "astro_options": AstroOptions(
                    USE_MINI_HALOS=True, INHOMO_RECO=True, USE_TS_FLUCT=True
                ),
            },
        ),
    ]

    def setup_class(self):
        """Create a default InputParameters."""
        self.default = InputParameters(random_seed=1)
        self.default_sigma8 = InputParameters(
            random_seed=1, cosmo_params=CosmoParams(SIGMA_8=1.0)
        )
        self.default_A_s = InputParameters(
            random_seed=1, cosmo_params=CosmoParams(A_s=3.0e-9)
        )

    @pytest.mark.parametrize(("exc", "msg", "kw"), EXCEPTION_CASES)
    def test_validation_exceptions(self, exc, msg, kw):
        """Test various exceptions that can happen on validation."""
        with config.use(ignore_R_BUBBLE_MAX_error=False), pytest.raises(exc, match=msg):
            InputParameters(random_seed=1, **kw)

    @pytest.mark.parametrize(("msg", "kw"), WARNINGS_CASES)
    def test_validation_warnings(self, msg, kw):
        """Test various warnings that can happen on validation."""
        with pytest.warns(UserWarning, match=msg):
            InputParameters(random_seed=1, **kw)

    def test_default(self):
        """Test the default object is, well, default."""
        assert self.default.cosmo_params == CosmoParams.new()
        assert self.default.simulation_options == SimulationOptions.new()
        assert self.default.matter_options == MatterOptions.new()
        assert self.default.astro_params == AstroParams.new()
        assert self.default.astro_options == AstroOptions.new()

    def test_evolve(self):
        """Test that evolve_input_structs does what it says."""
        altered_struct = self.default.evolve_input_structs(BOX_LEN=100)
        assert altered_struct.simulation_options.BOX_LEN == 100

        assert self.default.cosmo_tables.ps_norm == self.default.cosmo_params.SIGMA_8
        assert self.default.cosmo_tables.USE_SIGMA_8

        altered_struct_CLASS = self.default.evolve_input_structs(
            POWER_SPECTRUM="CLASS", K_MAX_FOR_CLASS=1.0
        )
        altered_struct2 = self.default.evolve_input_structs(OMm=0.5)
        altered_struct3 = altered_struct2.evolve_input_structs(POWER_SPECTRUM="EH")

        assert self.default.cosmo_tables.transfer_density is None
        assert altered_struct_CLASS.cosmo_tables.transfer_density is not None
        assert altered_struct_CLASS.cosmo_tables != altered_struct2.cosmo_tables
        assert altered_struct3.cosmo_tables.transfer_density is None
        assert (
            altered_struct_CLASS.cosmo_tables.ps_norm
            == altered_struct_CLASS.cosmo_params.SIGMA_8
        )
        assert altered_struct_CLASS.cosmo_tables.USE_SIGMA_8

        altered_struct = self.default.evolve_input_structs(SIGMA_8=1.0)
        assert altered_struct.cosmo_params.SIGMA_8 == 1.0

        altered_struct = self.default.evolve_input_structs(A_s=3.0e-9)
        assert altered_struct.cosmo_params.A_s == 3.0e-9
        # Even though we work with A_s, transfer function is EH, so we still use sigma8 at the backend
        assert (
            altered_struct.cosmo_tables.ps_norm == altered_struct.cosmo_params.SIGMA_8
        )
        assert altered_struct.cosmo_tables.USE_SIGMA_8

        # Test defaults with kwargs
        altered_struct = self.default_sigma8.evolve_input_structs(SIGMA_8=1.0)
        assert altered_struct.cosmo_params.SIGMA_8 == 1.0
        assert (
            altered_struct.cosmo_tables.ps_norm == altered_struct.cosmo_params.SIGMA_8
        )
        assert altered_struct.cosmo_tables.USE_SIGMA_8

        altered_struct = self.default_A_s.evolve_input_structs(A_s=3.0e-9)
        assert altered_struct.cosmo_params.A_s == 3.0e-9
        # Even though we work with A_s, transfer function is EH, so we still use sigma8 at the backend
        assert (
            altered_struct.cosmo_tables.ps_norm == altered_struct.cosmo_params.SIGMA_8
        )
        assert altered_struct.cosmo_tables.USE_SIGMA_8

        # Check that we can set A_s over the CLASS inputs and that we actually use it at the backend
        altered_struct = altered_struct_CLASS.evolve_input_structs(A_s=3.0e-9)
        assert altered_struct.cosmo_tables.ps_norm == altered_struct.cosmo_params.A_s
        assert not altered_struct.cosmo_tables.USE_SIGMA_8
        # Now check that if we switch to sigma8, we use it at the backend!
        altered_struct = altered_struct_CLASS.evolve_input_structs(
            SIGMA_8=1.0, A_s=None
        )
        assert (
            altered_struct.cosmo_tables.ps_norm == altered_struct.cosmo_params.SIGMA_8
        )
        assert altered_struct.cosmo_tables.USE_SIGMA_8
        # Last check, switching agaim from sigma8 to A_s (but this time SIGMA_8 is not None)
        altered_struct = altered_struct_CLASS.evolve_input_structs(
            A_s=3.0e-9, SIGMA_8=None
        )
        assert altered_struct.cosmo_tables.ps_norm == altered_struct.cosmo_params.A_s
        assert not altered_struct.cosmo_tables.USE_SIGMA_8

        # Check that we cannot work with both parameters
        with pytest.raises(ValueError, match="Cannot set both SIGMA_8 and A_s!"):
            self.default_sigma8.evolve_input_structs(A_s=3.0e-9)

        with pytest.raises(ValueError, match="Cannot set both SIGMA_8 and A_s!"):
            self.default_A_s.evolve_input_structs(SIGMA_8=1.0)

        # Check that we can change normalization parameter if we set the other parameter to None
        altered_struct = self.default_sigma8.evolve_input_structs(
            A_s=3.0e-9, SIGMA_8=None
        )
        assert altered_struct.cosmo_params.A_s == 3.0e-9
        # Even though we work with A_s, transfer function is EH, so we still use sigma8 at the backend
        assert (
            altered_struct.cosmo_tables.ps_norm == altered_struct.cosmo_params.SIGMA_8
        )
        assert altered_struct.cosmo_tables.USE_SIGMA_8
        altered_struct = self.default_A_s.evolve_input_structs(SIGMA_8=1.0, A_s=None)
        assert altered_struct.cosmo_params.SIGMA_8 == 1.0
        assert (
            altered_struct.cosmo_tables.ps_norm == altered_struct.cosmo_params.SIGMA_8
        )
        assert altered_struct.cosmo_tables.USE_SIGMA_8

    @pytest.mark.parametrize("template", _ALL_ALIASES)
    def test_from_template(self, template):
        """Test that creation from a template works for all templates."""
        # Some templates require running CLASS. We set K_MAX_FOR_CLASS to be small so the test won't take too long
        inputs = InputParameters.from_template(
            template, random_seed=1, K_MAX_FOR_CLASS=1.0
        )
        assert isinstance(inputs, InputParameters)

    def test_bad_input(self):
        """Test that passing a non-existent parameter to evolve raises."""
        with pytest.raises(
            TypeError,
            match="BAD_INPUT is not a valid keyword input",
        ):
            InputParameters(random_seed=0).evolve_input_structs(BAD_INPUT=True)

    def test_halomass_ranges(self):
        """Test that passing a non-existent parameter to evolve raises."""
        with pytest.raises(
            ValueError,
            match="There is a gap/overlap in the halo mass ranges",
        ):
            # These cells are ~6e7 Msun, with 1e8 minimum sampler mass this leaves a gap
            self.default.evolve_input_structs(BOX_LEN=30.0)

        with pytest.warns(
            UserWarning,
            match="The maximum halo mass",
        ):
            # The cell size is ~1e11 Msun
            self.default.evolve_input_structs(
                SOURCE_MODEL="L-INTEGRAL",
                USE_UPPER_STELLAR_TURNOVER=False,
            )



"""Testing plots is kind of hard, but we just check that it runs through without crashing."""

import pytest

import py21cmfast as p21c
from py21cmfast import plotting


def test_coeval_sliceplot(ic: p21c.InitialConditions):
    fig, ax = plotting.coeval_sliceplot(ic)

    assert ax.xaxis.get_label().get_text() == "x-axis [Mpc]"
    assert ax.yaxis.get_label().get_text() == "y-axis [Mpc]"

    with pytest.raises(
        ValueError, match="slice_axis should be between -1 and 2"
    ):  # bad slice axis
        plotting.coeval_sliceplot(ic, slice_axis=-2)

    with pytest.raises(
        IndexError, match="slice_index is too large"
    ):  # tring to plot slice that doesn't exist
        plotting.coeval_sliceplot(ic, slice_index=50)

    fig2, ax2 = plotting.coeval_sliceplot(ic, fig=fig, ax=ax)

    assert fig2 is fig
    assert ax2 is ax

    fig2, ax2 = plotting.coeval_sliceplot(ic, fig=fig)

    assert fig2 is fig
    assert ax2 is ax

    fig2, ax2 = plotting.coeval_sliceplot(ic, ax=ax)

    assert fig2 is fig
    assert ax2 is ax

    fig, ax = plotting.coeval_sliceplot(
        ic, kind="hires_density", slice_index=50, slice_axis=1
    )
    assert ax.xaxis.get_label().get_text() == "x-axis [Mpc]"
    assert ax.yaxis.get_label().get_text() == "z-axis [Mpc]"


def test_lightcone_sliceplot_default(lc: p21c.LightCone):
    _fig, ax = plotting.lightcone_sliceplot(lc)

    assert ax.yaxis.get_label().get_text() == "y-axis [Mpc]"
    assert ax.xaxis.get_label().get_text() == "Redshift"


def test_lightcone_sliceplot_vertical(lc: p21c.LightCone):
    _fig, ax = plotting.lightcone_sliceplot(lc, vertical=True)

    assert ax.yaxis.get_label().get_text() == "Redshift"
    assert ax.xaxis.get_label().get_text() == "y-axis [Mpc]"


def test_lc_sliceplot_freq(lc: p21c.LightCone):
    _fig, ax = plotting.lightcone_sliceplot(lc, zticks="frequency")

    assert ax.yaxis.get_label().get_text() == "y-axis [Mpc]"
    assert ax.xaxis.get_label().get_text() == "Frequency [MHz]"


def test_lc_sliceplot_cdist(lc: p21c.LightCone):
    _fig, ax = plotting.lightcone_sliceplot(lc, zticks="comoving_distance")

    assert ax.yaxis.get_label().get_text() == "y-axis [Mpc]"
    assert ax.xaxis.get_label().get_text() == "Comoving Distance [Mpc]"

    xlim = ax.get_xticks()
    assert xlim.min() >= 0
    assert xlim.max() <= lc.lightcone_dimensions[-1]


def test_lc_sliceplot_sliceax(lc: p21c.LightCone):
    _fig, ax = plotting.lightcone_sliceplot(lc, slice_axis=2)

    assert ax.yaxis.get_label().get_text() == "y-axis [Mpc]"
    assert ax.xaxis.get_label().get_text() == "x-axis [Mpc]"


def test_global_plot(lc: p21c.LightCone):
    _fig, ax = plotting.plot_global_history(lc)

    assert ax.xaxis.get_label().get_text() == "Redshift"



"""Run a quick simulation to test memory leaks."""

import os
import tracemalloc
from collections import deque

import psutil

import py21cmfast as p21c

tracemalloc.start()
snapshot = tracemalloc.take_snapshot()
PROCESS = psutil.Process(os.getpid())
oldmem = 0


def trace_print():
    """Print a trace of memory leaks."""
    global snapshot
    global oldmem

    snapshot2 = tracemalloc.take_snapshot()
    snapshot2 = snapshot2.filter_traces(
        (
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
            tracemalloc.Filter(False, tracemalloc.__file__),
        )
    )

    if snapshot is not None:
        thismem = PROCESS.memory_info().rss / 1024**2
        diff = thismem - oldmem
        print(
            f"===================== Begin Trace (TOTAL MEM={thismem:1.4e} MB... [{diff:+1.4e} MB]):"
        )
        top_stats = snapshot2.compare_to(snapshot, "lineno", cumulative=True)
        for stat in top_stats[:4]:
            print(stat)
        print("End Trace ===========================================")
        oldmem = thismem

    snapshot = snapshot2


inputs = p21c.InputParameters.from_template(
    ["Munoz21", "medium"], random_seed=0, MINIMIZE_MEMORY=True, ZPRIME_STEP_FACTOR=1.2
)

print(inputs)

coevals = deque(maxlen=3)

for coeval, _ in p21c.generate_coeval(
    inputs=inputs, cache=p21c.OutputCache("EvolutionMemoryLeakCache"), write=True
):
    print(coeval.redshift)
    trace_print()

    print("---------------")
    coevals.append(coeval)
    for ostruct in coevals[0].output_structs.values():
        print(ostruct.__class__.__name__)
        for name, ary in ostruct.arrays.items():
            print(f"   {name}: {ary.state}")
    print("---------------")



"""Test the GlobalEvolution class."""

from pathlib import Path

import numpy as np
import pytest

import py21cmfast as p21c
from py21cmfast import GlobalEvolution

DATA_PATH = Path(__file__).parent / "test_data"


@pytest.mark.parametrize("source_model", ["CONST-ION-EFF", "E-INTEGRAL", "L-INTEGRAL"])
def test_global_quantities(default_input_struct_ts, source_model):
    """Test that global quantities behave as expected."""
    global_evolution = p21c.run_global_evolution(
        inputs=default_input_struct_ts.with_logspaced_redshifts(),
        source_model=source_model,
    )

    z = np.array(global_evolution.node_redshifts)
    T_gamma = 2.7255 * (1.0 + z)
    T_k = global_evolution.quantities["kinetic_temp_neutral"]
    T_s = global_evolution.quantities["spin_temperature"]
    x_HI = global_evolution.quantities["neutral_fraction"]
    T_21 = global_evolution.quantities["brightness_temp"]

    # Find where the minimum temperature is
    min_idx_Tk = np.argmin(T_k)
    min_idx_Ts = np.argmin(T_s)
    # Check it's not at the edges (i.e., there's a true minimum)
    assert 0 < min_idx_Tk < len(T_k) - 1
    assert 0 < min_idx_Ts < len(T_s) - 1
    # Test that the gas is cooled down (by adiabatic cooling) before it is heated up (by X-rays)
    # NOTE: we don't check that the spin temperature continues to climb after the minimum,
    #       because the Lya coupling could decrease at some point, thus lowering the spin temperature
    assert np.all(np.diff(T_k[: min_idx_Tk + 1]) <= 0)
    assert np.all(np.diff(T_k[min_idx_Tk:]) >= 0)
    assert np.all(np.diff(T_s[: min_idx_Ts + 1]) <= 0)

    # Check that the spin temperature is contained between the gas temperature and the CMB temperature
    assert np.all(T_s <= np.maximum(T_gamma, T_k))
    assert np.all(np.minimum(T_gamma, T_k) <= T_s)

    # Make sure that the x_HI curve is montonotonously decreasing
    assert np.all(np.diff(x_HI) <= 0)

    # Test that global signal has two local maxima and minima
    sign_changes = np.diff(np.sign(np.diff(T_21)))
    local_maxima_indices = np.where(sign_changes < 0)[0]
    local_minima_indices = np.where(sign_changes > 0)[0]
    assert len(local_maxima_indices) == 2  # Lya coupling + reionization start
    assert len(local_minima_indices) == 2  # X-ray heating + reionization end

    # Check that the brightness temperature vanishes when reionization ended
    assert np.all(T_21[local_minima_indices[1] + 1 :] == 0.0)
    assert np.all(x_HI[local_minima_indices[1] + 1 :] == 0.0)


@pytest.mark.parametrize("source_model", ["CONST-ION-EFF", "E-INTEGRAL", "L-INTEGRAL"])
def test_run_global_evolution_without_source_model(
    default_input_struct_ts, source_model
):
    """Test that run_global_evolution works without specifying source_model."""
    global_evolution = p21c.run_global_evolution(
        inputs=default_input_struct_ts.evolve_input_structs(SOURCE_MODEL=source_model)
    )
    assert isinstance(global_evolution, GlobalEvolution)


def test_run_global_evolution_without_Ts(default_input_struct):
    """Test that run_global_evolution doesn't crash when USE_TS_FLUCT=False."""
    global_evolution = p21c.run_global_evolution(
        inputs=default_input_struct.evolve_input_structs(USE_TS_FLUCT=False)
    )
    assert isinstance(global_evolution, GlobalEvolution)


def test_run_global_evolution_from_template():
    """Test that run_global_evolution doesn't crash when using a template."""
    global_evolution = p21c.run_global_evolution(
        inputs=p21c.InputParameters.from_template("simple", random_seed=1234)
    )
    assert isinstance(global_evolution, GlobalEvolution)


def test_global_evolution_roundtrip(test_direc, default_input_struct_ts):
    """Test that the save/from_file methods yield the same GlobalEvolution object."""
    global_evolution1 = p21c.run_global_evolution(inputs=default_input_struct_ts)
    fname = test_direc / "global_evolution.h5"
    global_evolution1.save(path=fname)
    global_evolution2 = GlobalEvolution.from_file(fname)

    assert global_evolution1 == global_evolution2


@pytest.mark.parametrize("source_model", ["X-INTEGRAL", "DEXM-ESF", "CHMF-SAMPLER"])
def test_global_evolution_bad_inputs(default_input_struct_ts, source_model):
    """Test run_global_evolution with bad inputs."""
    with pytest.raises(ValueError, match="'source_model' must be one of"):
        p21c.run_global_evolution(
            inputs=default_input_struct_ts, source_model=source_model
        )
    if source_model == "X-INTEGRAL":
        pytest.xfail("X-INTEGRAL is not a valid option for SOURCE_MODEL.")
    else:
        with pytest.raises(
            ValueError,
            match="You did not specify 'source_model', but SOURCE_MODEL in `inputs` has discrete halos!",
        ):
            p21c.run_global_evolution(
                inputs=default_input_struct_ts.evolve_input_structs(
                    SOURCE_MODEL=source_model
                ),
            )


def test_compatability_with_database():
    """
    Test that loading GlobalEvolution from the database succeeds.

    This test is useful for checking compatability with data files that
    were generated with older versions of 21cmFAST.

    If this test fails DO NOT replace the file in the database, unless you
    are absolutely sure of what you are doing.
    """
    fname = DATA_PATH / "global_evolution.h5"
    global_evolution = GlobalEvolution.from_file(fname)
    assert isinstance(global_evolution, GlobalEvolution)


def test_linear_perturbation_theory(default_input_struct_ts):
    """
    Test that we can do linear perturbation theory with 21cmFAST.

    What this test does is to confirm that the contrast of the fields we compute does not depend
    on our chosen initial perturbation, at first order.
    """
    delta1 = 1e-7
    delta2 = 1e-8

    # Compute global evolution three times.
    # We stop at relatively high redshift, where linear perturbation theory becomes non-valid below it
    inputs = default_input_struct_ts.with_logspaced_redshifts(zmin=20)
    global_evolution_0 = p21c.run_global_evolution(inputs=inputs)
    global_evolution_delta1 = p21c.run_global_evolution(
        inputs=inputs, overdensity_z0=delta1
    )
    global_evolution_delta2 = p21c.run_global_evolution(
        inputs=inputs, overdensity_z0=delta2
    )

    for quantity in global_evolution_0.quantities:
        f_0 = global_evolution_0.quantities[quantity]

        f_delta1 = global_evolution_delta1.quantities[quantity]
        contrast1 = (f_delta1 - f_0) / delta1

        f_delta2 = global_evolution_delta2.quantities[quantity]
        contrast2 = (f_delta2 - f_0) / delta2

        np.testing.assert_allclose(contrast1, contrast2, atol=0, rtol=1e-5)


def test_linear_density_field(default_input_struct):
    """Test that the linear density field grows linearly with time."""
    delta1 = 1e-7
    delta2 = 1e-8

    global_evolution1 = p21c.run_global_evolution(
        inputs=default_input_struct, overdensity_z0=delta1
    )
    global_evolution2 = p21c.run_global_evolution(
        inputs=default_input_struct, overdensity_z0=delta2
    )

    growth_factor1 = global_evolution1.quantities["density"] / delta1
    growth_factor2 = global_evolution2.quantities["density"] / delta2

    # Test that the linear density field is linear
    np.testing.assert_allclose(growth_factor1, growth_factor2, atol=0, rtol=1e-5)

    # Test that the linear density field grows with time
    assert np.all(np.diff(growth_factor1) < 0)



"""Tests of switching on minimize memory flag."""

import numpy as np
import pytest

import py21cmfast as p21c


@pytest.mark.parametrize("source_model", ["L-INTEGRAL", "E-INTEGRAL"])
def test_minimize_memory_on_global_evolution(source_model: str):
    """Test that switching minimize memory on/off doesn't change outputs.

    We test this for two source models: one on the lagrangian grid (L-INTEGRAL) and one
    on the Eulerian grid (E-INTEGRAL). The former should not be affected at all by the
    MINIMIZE_MEMORY flag, while the latter should be significantly affected in terms of
    memory usage, but not in terms of results. We test this by running the global
    evolution for both cases and checking that the brightness temperature evolution is
    unchanged.
    """
    inputs = p21c.InputParameters.from_template(
        ["park19", "tiny"],
        random_seed=1234,
        MINIMIZE_MEMORY=False,
        SOURCE_MODEL=source_model,
    )
    inputs_minmem = inputs.evolve_input_structs(MINIMIZE_MEMORY=True)

    global_evolution = p21c.run_global_evolution(inputs=inputs)
    global_evolution_minmem = p21c.run_global_evolution(inputs=inputs_minmem)

    np.testing.assert_allclose(
        global_evolution.quantities["brightness_temp"],
        global_evolution_minmem.quantities["brightness_temp"],
        atol=0.1,  # mK
    )



"""Open and read the configuration file."""

from __future__ import annotations

import contextlib
from pathlib import Path
from typing import ClassVar

from . import yaml
from ._data import DATA_PATH
from .c_21cmfast import ffi, lib
from .wrapper.structs import StructInstanceWrapper


class ConfigurationError(Exception):
    """An error with the config file."""


# TODO: force config to be a singleton
class Config(dict):
    """Simple over-ride of dict that adds a context manager."""

    _defaults: ClassVar = {
        "direc": "~/21cmFAST-cache",
        "ignore_R_BUBBLE_MAX_error": False,
        "external_table_path": DATA_PATH,
        "HALO_CATALOG_MEM_FACTOR": 1.2,
        "EXTRA_HALOBOX_FIELDS": False,
        "safe_read": True,
    }
    _defaults["wisdoms_path"] = Path(_defaults["direc"]) / "wisdoms"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # keep the config settings from the C library here
        self._c_config_settings = StructInstanceWrapper(lib.config_settings, ffi)

        for k, v in self._defaults.items():
            if k not in self:
                self[k] = v

        for k in self.keys():
            if k not in self._defaults:
                raise ConfigurationError(
                    f"You passed the key '{k}' to config, which is not known to 21cmFAST."
                )

        self["direc"] = Path(self["direc"]).expanduser().absolute()

        # since the subclass __setitem__ is not called in the super().__init__ call, we re-do the setting here
        # NOTE: This seems messy but I don't know a better way to do it
        for k in self._c_config_settings:
            self._pass_to_backend(k, self[k])

    def __setitem__(self, key, value):
        """Set an item in the config. Also updating the backend if it exists there."""
        super().__setitem__(key, value)
        if key in self._c_config_settings:
            self._pass_to_backend(key, value)

    def _pass_to_backend(self, key, value):
        """Set the value in the backend."""
        # we should possibly do a typemap for the ffi
        if isinstance(value, Path | str):
            setattr(
                self._c_config_settings, key, ffi.new("char[]", str(value).encode())
            )
        else:
            setattr(self._c_config_settings, key, value)

    @contextlib.contextmanager
    def use(self, **kwargs):
        """Context manager for using certain configuration options for a set time."""
        backup = self.copy()
        for k, v in kwargs.items():
            self[k] = Path(v).expanduser().absolute() if k == "direc" else v
        yield self
        for k in kwargs:
            self[k] = backup[k]

    def write(self, fname: str | Path | None = None):
        """Write current configuration to file to make it permanent."""
        if fname := Path(fname or self.file_name):
            if not fname.parent.exists():
                fname.parent.mkdir(parents=True)

            with fname.open("w") as fl:
                yaml.dump(self._as_dict(), fl)

    def _as_dict(self):
        """Return a plain dict defining the instance."""
        return {k: str(Path) if isinstance(v, Path) else v for k, v in self.items()}

    @classmethod
    def load(cls, file_name: str | Path):
        """Create a Config object from a config file."""
        file_name = Path(file_name).expanduser().absolute()

        if file_name.exists():
            with file_name.open() as fl:
                cfg = yaml.load(fl)
            return cls(cfg, file_name=file_name)
        else:
            return cls(write=True)


# On import, load the default config
config = Config()



"""Tests of photon conservation in 21cmFAST."""

import pytest

import py21cmfast as p21c


@pytest.mark.parametrize(
    "model",
    [
        "no-photoncons",
        "z-photoncons",
        "alpha-photoncons",
        "f-photoncons",
    ],
)
def test_memory_accesss(model, tiny_inputs, tiny_ics):
    """Simply tests that no segfaults occur."""
    inputs = tiny_inputs.evolve_input_structs(PHOTON_CONS_TYPE=model)

    p21c.run_coeval(
        inputs=inputs, initial_conditions=tiny_ics, write=False, progressbar=True
    )



"""
Unit-tests of the lightcone drivers.

They do not test for correctness of simulations, but whether different parameter
options work/don't work as intended.
"""

import numpy as np
import pytest

import py21cmfast as p21c


def test_lightcone(
    lc, default_simulation_options, lightcone_min_redshift, max_redshift
):
    assert lc.lightcone_redshifts[-1] >= max_redshift
    assert np.isclose(lc.lightcone_redshifts[0], lightcone_min_redshift, atol=1e-4)
    assert (
        lc.cell_size
        == default_simulation_options.BOX_LEN / default_simulation_options.HII_DIM
    )


def test_lightcone_quantities(
    ic, default_input_struct_lc, lightcone_min_redshift, max_redshift, cache
):
    lcn = p21c.RectilinearLightconer.with_equal_cdist_slices(
        min_redshift=lightcone_min_redshift,
        max_redshift=max_redshift,
        resolution=ic.simulation_options.cell_size,
        cosmo=ic.cosmo_params.cosmo,
        quantities=(
            "density",
            "brightness_temp",
            "ionisation_rate_G12",
        ),
    )

    lc = p21c.run_lightcone(
        lightconer=lcn,
        initial_conditions=ic,
        inputs=default_input_struct_lc,
        cache=cache,
    )

    assert "density" in lc.lightcones
    assert "brightness_temp" in lc.lightcones
    assert "ionisation_rate_G12" in lc.lightcones
    assert "ionisation_rate_G12" in lc.global_quantities
    assert "density" in lc.global_quantities
    assert "log10_mturn_acg" in lc.global_quantities
    assert "log10_mturn_mcg" in lc.global_quantities

    # density should be filled with not zeros.
    assert lc.lightcones["density"].min() != lc.lightcones["density"].max() != 0

    # Simply ensure that different quantities are not getting crossed/referred to each other.
    assert (
        lc.lightcones["density"].min()
        != lc.lightcones["brightness_temp"].min()
        != lc.lightcones["brightness_temp"].max()
    )

    lcn_ts = p21c.RectilinearLightconer.with_equal_cdist_slices(
        min_redshift=lightcone_min_redshift,
        max_redshift=max_redshift,
        resolution=ic.simulation_options.cell_size,
        cosmo=ic.cosmo_params.cosmo,
        quantities=("spin_temperature", "density"),
    )

    # Raise an error since we're not doing spin temp.
    with pytest.raises(AttributeError):
        p21c.run_lightcone(
            lightconer=lcn_ts,
            initial_conditions=ic,
            inputs=default_input_struct_lc,
            cache=cache,
        )


def test_lightcone_coords(lc):
    assert lc.lightcone_coords.shape == (lc.lightcone_distances.shape[0],)
    assert lc.lightcone_coords[0] == 0.0
    np.testing.assert_allclose(
        np.diff(lc.lightcone_coords.to_value("Mpc")),
        lc.simulation_options.BOX_LEN / lc.simulation_options.HII_DIM,
    )


def test_run_lc_bad_inputs(
    rectlcn,
    default_input_struct_lc: p21c.InputParameters,
    cache,
):
    with pytest.raises(
        ValueError,
        match="You are attempting to run a lightcone with no node_redshifts",
    ):
        p21c.run_lightcone(
            lightconer=rectlcn,
            inputs=default_input_struct_lc.clone(node_redshifts=[]),
        )


def test_lc_with_lightcone_filename(
    ic, rectlcn, default_input_struct_lc, tmpdirec, cache
):
    fname = tmpdirec / "lightcone.h5"
    lc = p21c.run_lightcone(
        lightconer=rectlcn,
        initial_conditions=ic,
        inputs=default_input_struct_lc,
        lightcone_filename=fname,
        cache=cache,
    )
    assert fname.exists()

    lc_loaded = p21c.LightCone.from_file(fname)
    assert lc_loaded == lc
    del lc_loaded

    # This one should NOT run anything.
    lc2 = p21c.run_lightcone(
        lightconer=rectlcn,
        initial_conditions=ic,
        inputs=default_input_struct_lc,
        lightcone_filename=fname,
        cache=cache,
    )

    assert lc2 == lc
    del lc2

    fname.unlink()


def test_lc_partial_eval(rectlcn, default_input_struct_lc, tmpdirec, lc, cache):
    fname = tmpdirec / "lightcone_partial.h5"

    z = rectlcn.lc_redshifts.max()
    lc_gen = p21c.generate_lightcone(
        lightconer=rectlcn,
        inputs=default_input_struct_lc,
        lightcone_filename=fname,
        write=True,
        cache=cache,
        include_dvdr_in_tau21=False,
    )
    while z > 20.0:
        iz_1, z, _, partial = next(lc_gen)

    assert 0 < partial._last_completed_node < len(rectlcn.lc_redshifts) - 1

    np.testing.assert_allclose(
        partial.global_quantities["brightness_temp"][
            : partial._last_completed_node + 1
        ],
        lc.global_quantities["brightness_temp"][: partial._last_completed_node + 1],
        rtol=1e-6,
    )
    np.testing.assert_allclose(
        partial.lightcones["brightness_temp"][partial._last_completed_lcidx :],
        lc.lightcones["brightness_temp"][partial._last_completed_lcidx :],
        rtol=1e-4,
    )

    lc_gen_cont = p21c.generate_lightcone(
        lightconer=rectlcn,
        inputs=default_input_struct_lc,
        lightcone_filename=fname,
        cache=cache,
        write=False,
        regenerate=False,
        include_dvdr_in_tau21=False,
    )
    for iz_2, z, _, finished in lc_gen_cont:  # noqa: B007
        assert z <= 20.0
        assert iz_2 > iz_1

    # this only checks the lightcone object, not the array fields
    assert finished == lc
    # make sure we wrote to the correct indices
    np.testing.assert_allclose(
        finished.global_quantities["brightness_temp"],
        lc.global_quantities["brightness_temp"],
        rtol=1e-6,
    )
    np.testing.assert_allclose(
        finished.lightcones["brightness_temp"],
        lc.lightcones["brightness_temp"],
        rtol=1e-4,
    )


def test_lc_lowerz_than_photon_cons(
    ic, default_input_struct_lc, default_astro_options, max_redshift, cache
):
    inputs = default_input_struct_lc.clone(
        node_redshifts=p21c.get_logspaced_redshifts(
            min_redshift=1.9,
            max_redshift=max(default_input_struct_lc.node_redshifts),
            z_step_factor=default_input_struct_lc.simulation_options.ZPRIME_STEP_FACTOR,
        ),
        astro_options=default_astro_options.clone(PHOTON_CONS_TYPE="z-photoncons"),
    )
    lcn = p21c.RectilinearLightconer.with_equal_cdist_slices(
        min_redshift=2.0,
        max_redshift=max_redshift,
        resolution=ic.simulation_options.cell_size,
        cosmo=ic.cosmo_params.cosmo,
    )

    with pytest.raises(ValueError, match="You have passed a redshift"):
        p21c.run_lightcone(
            lightconer=lcn, initial_conditions=ic, inputs=inputs, cache=cache
        )



"""Module containing a driver function for computing the global evolution of the fields in the simulation."""

import warnings
from pathlib import Path
from typing import Self

import attrs
import h5py
import numpy as np

from .. import __version__
from ..c_21cmfast import lib
from ..io import h5
from ..io.caching import CacheConfig
from ..wrapper.arrays import Array
from ..wrapper.cfuncs import evaluate_Nion_z
from ..wrapper.inputs import InputParameters
from ..wrapper.outputs import (
    BrightnessTemp,
    HaloBox,
    IonizedBox,
    TsBox,
)
from .coeval import _redshift_loop_generator, _setup_ics_and_pfs_for_scrolling


def compute_global_reionization_at_z(
    redshift: float,
    inputs: InputParameters,
    previous_ionized_box: IonizedBox,
    spin_temp: TsBox | None,
) -> IonizedBox:
    r"""
    Compute an ionized box at a given redshift, according to the global evolution.

    Parameters
    ----------
    inputs : :class:`~InputParameters`
        The input parameters specifying the run. Since this may be the first box
        to use the astro params/flags, it is needed when we have not computed a TsBox or HaloBox.
    previous_ionize_box: :class:`IonizedBox`
        An ionized box at higher redshift.
    spin_temp: :class:`TsBox` or None, optional
        A spin-temperature box, only required if `USE_TS_FLUCT` is True. This is useful since
        the volume filling factor Q_HI is already contained in this object. If None, it will be
        evaluated by calling the C code.

    Returns
    -------
    :class:`~IonizedBox` :
        An object containing the ionized box data.
    """
    box = IonizedBox.new(inputs=inputs, redshift=redshift)
    shape = (1, 1, 1)

    if spin_temp is None:
        # We compute Q_HI very similarly as in SpinTemperatureBox.c.
        nion, _ = evaluate_Nion_z(
            inputs=inputs,
            redshifts=np.asarray(redshift),
            # I just put here an arbitrary number because we currently don't allow to have mini-halos when USE_TS_FLUCT=False.
            # TODO: should this limitation be relaxed in the future?
            log10mturns=np.asarray(5.0),
        )
        if inputs.matter_options.SOURCE_MODEL == "CONST-ION-EFF":
            ion_eff_factor = inputs.astro_params.HII_EFF_FACTOR
        else:
            ion_eff_factor = (
                pow(10.0, inputs.astro_params.F_STAR10)
                * pow(10.0, inputs.astro_params.F_ESC10)
                * inputs.astro_params.POP2_ION
            )
        Q_HI = 1.0 - ion_eff_factor * nion
    else:
        Q_HI = spin_temp.Q_HI

    # TODO: I think a more accurate global Q_HI can be achieved by solving an ODE that includes also the recombination rate
    Q_HI = Q_HI if Q_HI > 0.0 else 0.0

    # A crude way to estimate the global photoionization rate
    try:
        dQdz = (Q_HI - previous_ionized_box.neutral_fraction.value) / (
            redshift - previous_ionized_box.redshift
        )
    except TypeError:
        dQdz = 0.0
    dzdt = -(1.0 + redshift) * inputs.cosmo_params.cosmo.H(redshift)
    ionisation_rate_G12 = np.abs(dQdz * dzdt)
    # TODO: is there a more clever way to estimate global z_reion?
    z_reion = -1.0 if Q_HI > 0.0 else redshift

    # Now initialize the output box with the global values!
    required_arrays = {
        "neutral_fraction": Q_HI,
        "ionisation_rate_G12": ionisation_rate_G12.to("1/s"),
        "z_reion": z_reion,
    }
    for name, val in required_arrays.items():
        setattr(
            box,
            name,
            Array(shape=shape, dtype=np.float32)
            .initialize()
            .with_value(val=val * np.ones(shape)),
        )
    return box


@attrs.define()
class GlobalEvolution:
    """Structure that contains the global evolution (as a function of redshift) of all the fields in the simulation.

    Attributes
    ----------
    inputs: InputParameters
        The input parameters for the simulation.
    quantities: dict[str, np.ndarray] | None
        Arrays of length `node_redshifts` containing the global field across redshift.
    """

    inputs: InputParameters = attrs.field(
        validator=attrs.validators.instance_of(InputParameters)
    )
    quantities: dict[str, np.ndarray] | None = attrs.field(default=None)

    @classmethod
    def get_fields(cls, inputs: InputParameters) -> tuple:
        """Get a list of the names of the available fields in the simulation."""
        possible_outputs = [
            BrightnessTemp.new(inputs, redshift=0),
        ]
        if inputs.astro_options.USE_TS_FLUCT:
            possible_outputs.append(TsBox.new(inputs, redshift=0))
        if inputs.matter_options.lagrangian_source_grid:
            possible_outputs.append(HaloBox.new(inputs, redshift=0))
        field_names = ("density", "neutral_fraction", "ionisation_rate_G12")
        for output in possible_outputs:
            field_names += tuple(output.arrays.keys())
        return field_names

    @property
    def simulation_options(self):
        """Matter params shared by all datasets."""
        return self.inputs.simulation_options

    @property
    def matter_options(self):
        """Matter flags shared by all datasets."""
        return self.inputs.matter_options

    @property
    def cosmo_params(self):
        """Cosmo params shared by all datasets."""
        return self.inputs.cosmo_params

    @property
    def astro_options(self):
        """Flag Options shared by all datasets."""
        return self.inputs.astro_options

    @property
    def astro_params(self):
        """Astro params shared by all datasets."""
        return self.inputs.astro_params

    @property
    def cosmo_tables(self):
        """Cosmo tables shared by all datasets."""
        return self.inputs.cosmo_tables

    @property
    def random_seed(self):
        """Random seed shared by all datasets."""
        return self.inputs.random_seed

    @property
    def node_redshifts(self):
        """Redshifts at which coeval boxes and global quantities are computed."""
        return self.inputs.node_redshifts

    def save(
        self,
        path: str | Path,
        clobber=False,
    ):
        """Save the global_evolution object to disk."""
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        file_mode = "w" if clobber else "a"
        with h5py.File(path, file_mode) as fl:
            fl.attrs["global_evolution"] = (
                True  # marker identifying this as a global_evolution
            )

            fl.attrs["__version__"] = __version__

            global_q = fl.create_group("quantities")
            for k, v in self.quantities.items():
                global_q[k] = v

        h5._write_inputs_to_group(self.inputs, path)

    @classmethod
    def from_file(
        cls, path: str | Path, safe: bool = True, remove_buffer: bool = True
    ) -> Self:
        """Create a new instance from a saved global_evolution on disk."""
        kwargs = {}
        with h5py.File(path, "r") as fl:
            if not fl.attrs.get("global_evolution", False):
                raise ValueError(f"The file {path} is not a global_evolution file!")

            kwargs["inputs"] = h5.read_inputs(fl, safe=safe)

            glb = fl["quantities"]
            kwargs["quantities"] = {k: glb[k][...] for k in glb}

        return cls(**kwargs)

    def __eq__(self, other):
        """Determine if this is equal to another object."""
        return (
            isinstance(other, self.__class__)
            and self.inputs == other.inputs
            and self.quantities.keys() == other.quantities.keys()
        )


def run_global_evolution(
    inputs: InputParameters,
    source_model: str | None = None,
    progressbar: bool = False,
    overdensity_z0: float | None = None,
):
    r"""
    Compute the global evolution of all the fields in the simulation.

    Parameters
    ----------
    inputs: :class:`~InputParameters`
        This object specifies the input parameters for the run, including the random seed.
    source_model: str, optional
        The source model to use in the simulation. If not provided, it is taken from inputs, unless inputs.matter_options.has_discrete_halos
        is True, in which case an error is thrown. Options are:
        E-INTEGRAL : The traditional excursion-set formalism, where source properties are
            defined on the Eulerian grid after 2LPT in regions of filter scale R (see the X_FILTER options for filter shapes).
            This integrates over the CHMF using the smoothed density grids, then multiplies the result.
            by (1 + delta) to get the source properties in each cell.
        CONST-ION-EFF: Similar to E-INTEGRAL, but ionizing efficiency is constant and does not depend on the halo mass
            (see Mesinger+ 2010).
        L-INTEGRAL : Analagous to the 'ESF-L' model described in Trac+22, where source properties
            are defined on the Lagrangian (IC) grid by integrating the CHMF prior to the IGM physics
            and then mapping properties to the Eulerian grid using 2LPT.
    progressbar: bool, optional
        If True, a progress bar will be displayed throughout the simulation. Defaults to False.
    overdensity_z0: float, optional
        The linear matter overdensity at z=0. Default is 0. This input argument could be useful for doing linear perturbation analysis with 21cmFAST.

    Returns
    -------
    global_evolution : :class:`~py21cmfast.GlobalEvolution`
        The object containing the global evolution of the fields in the simulation.

    Raises
    ------
    ValueError: If source_model is None and inputs.matter_options.has_discrete_halos, or if source_model is provided,
    but is not one of the said options above.

    Notes
    -----
    For convenience, the provided InputParameters object by the user is overwritten to allow this function to work smoothly
    in all cases. The InputParameters object that was used in this function can be accessed via global_evolution.inputs.
    Note for example that the actual InputParameters object that is used by this function has only one cell (HII_DIM=DIM=1)
    and no 2LPT is performed (since we have only one cell).
    In addition, note that the classic reionization module of 21cmFAST (where the excursion-set algorithm is applied to find ionized
    bubbles) is not called by this function, since there are no bubbles in a one cell box! Instead, we compute the global reionization
    history based on the global volume filling factor Q. For that reason, note that the correction due to photon non-conservation in the
    excursion-set algorithm is never applied when this function is called.

    """
    # When doing glboal evolution, we allow only source models free of discrete halos
    possible_sources = ["CONST-ION-EFF", "E-INTEGRAL", "L-INTEGRAL"]

    if source_model is None:
        if inputs.matter_options.has_discrete_halos:
            raise ValueError(
                "You did not specify 'source_model', but SOURCE_MODEL in `inputs` has discrete halos! "
                "Either specify 'source_model' or change SOURCE_MODEL in `inputs` to a model with no discrete halos."
            )
        else:
            source_model = inputs.matter_options.SOURCE_MODEL

    if source_model not in possible_sources:
        raise ValueError(
            f"'source_model' must be one of {possible_sources}, got {source_model} instead."
        )

    if not inputs.astro_options.USE_TS_FLUCT:
        warnings.warn(
            "Your inputs.astro_options.USE_TS_FLUCT = False. "
            "While this is the default settings (to reduce computation time when full coeval boxes are evaluated), "
            "it yields the incorrect 21-cm signal at high redshifts, before the spin temperature saturates the "
            "background photon temperature. Consider changing USE_TS_FLUCT to True in order to get the correct "
            "global 21-cm signal at all redshifts.",
            stacklevel=2,
        )

    new_input_kwargs = {
        "HIRES_TO_LOWRES_FACTOR": None,  # we set below DIM=HII_DIM=1
        "DIM": 1,  # we need only one cell
        "HII_DIM": 1,  # we need only one cell
        "BOX_LEN": 1e6,  # we need a huge box/cell in order to simulate the global evolution
        "SOURCE_MODEL": source_model,
        "PERTURB_ALGORITHM": "LINEAR",  # no need to do 2LPT
        "USE_INTERPOLATION_TABLES": "sigma-interpolation",  # only need sigma interpolation tables (hmf integrals are evaluated once per snapshot, without interpolation)
        "INTEGRATION_METHOD_ATOMIC": "GSL-QAG",  # due to above, we ought to use gsl, and not gauss-legendre (BUG?)
        "INTEGRATION_METHOD_MINI": "GSL-QAG",
        "USE_UPPER_STELLAR_TURNOVER": False,  # no upper stellar turnover without discrete halos
        "USE_EXP_FILTER": False,  # we don't run reionization module, so we can leave this parameter on False for all source models
        "KEEP_3D_VELOCITIES": False,  # we don't need any velocities
        "PHOTON_CONS_TYPE": "no-photoncons",  # we don't do photon conservation
    }
    inputs_one_cell = inputs.evolve_input_structs(**new_input_kwargs)

    global_evolution = GlobalEvolution(
        inputs=inputs_one_cell,
        quantities={},
    )
    for quantity in global_evolution.get_fields(inputs_one_cell):
        global_evolution.quantities[quantity] = np.zeros(
            len(inputs_one_cell.node_redshifts)
        )

    prev_coeval = None

    iokw = {"cache": None, "regenerate": True, "free_cosmo_tables": False}

    (
        initial_conditions,
        perturbed_fields,
        halofield_list,
        photon_nonconservation_data,
    ) = _setup_ics_and_pfs_for_scrolling(
        all_redshifts=inputs_one_cell.node_redshifts,
        inputs=inputs_one_cell,
        initial_conditions=None,
        write=CacheConfig.off(),
        progressbar=progressbar,
        overdensity_z0=overdensity_z0,
        **iokw,
    )

    for iz, coeval in _redshift_loop_generator(
        inputs=inputs_one_cell,
        initial_conditions=initial_conditions,
        all_redshifts=inputs_one_cell.node_redshifts,
        perturbed_field=perturbed_fields,
        halofield_list=halofield_list,
        write=CacheConfig.off(),
        cleanup=True,
        progressbar=progressbar,
        photon_nonconservation_data=photon_nonconservation_data,
        init_coeval=prev_coeval,
        iokw=iokw,
    ):
        for quantity in global_evolution.quantities:
            global_evolution.quantities[quantity][iz] = np.mean(
                getattr(coeval, quantity)
            )

        prev_coeval = coeval

    lib.Free_cosmo_tables_global()

    return global_evolution



"""
Compute single physical fields.

These functions are high-level wrappers around C-functions that compute 3D fields, for
example initial conditions, perturbed fields and ionization fields.
"""

import logging
import warnings

import numpy as np
from astropy import units as un
from astropy.cosmology import z_at_value

from ..wrapper.arrays import Array
from ..wrapper.inputs import InputParameters
from ..wrapper.outputs import (
    BrightnessTemp,
    HaloBox,
    HaloCatalog,
    InitialConditions,
    IonizedBox,
    PerturbedField,
    PerturbedHaloCatalog,
    TsBox,
    XraySourceBox,
)
from ._param_config import (
    check_output_consistency,
    single_field_func,
)

logger = logging.getLogger(__name__)


@single_field_func
def compute_initial_conditions(
    *, inputs: InputParameters, initial_density: np.ndarray | float | None = None
) -> InitialConditions:
    r"""
    Compute initial conditions.

    Parameters
    ----------
    inputs
        The InputParameters instance defining the run.
    initial_density: np.ndarray or float, optional
        A realization of the density field on the high resolution grid.
        This input can also be used to determine the global density field,
        in case we have a single cell in the box.
    regenerate : bool, optional
        Whether to force regeneration of data, even if matching cached data is found.
    cache
        An OutputCache object defining how to read cached boxes.
    write
        A boolean specifying whether we need to cache the box.

    Returns
    -------
    :class:`~InitialConditions`
    """
    # Initialize memory for the boxes that will be returned.
    ics = InitialConditions.new(inputs=inputs)

    if inputs.simulation_options.HII_DIM == 1 and inputs.simulation_options.DIM == 1:
        # If we have only one cell (could happen if we do a global evolution), we don't really need to compute the ics
        shape = (1, 1, 1)
        required_arrays = PerturbedField.new(
            redshift=0, inputs=inputs
        ).get_required_input_arrays(ics)

        # Set the arrays to zero, or according to initial_density if the arrays are density fields
        for array in required_arrays:
            if (initial_density is not None) and (
                array in ["hires_density", "lowres_density"]
            ):
                value = initial_density
            else:
                value = 0.0
            setattr(
                ics,
                array,
                Array(shape=shape, dtype=np.float32)
                .initialize()
                .with_value(val=value * np.ones(shape)),
            )
        return ics
    else:
        if initial_density is not None:
            if np.abs(initial_density.mean() > 1e-3):
                warnings.warn(
                    f"Your initial_density has mean {initial_density.mean()}. "
                    + "Make sure you know what you are doing.",
                    stacklevel=2,
                )
            shape = ics.hires_density.shape
            if initial_density.shape != shape:
                raise ValueError(
                    "The shape of your high resolution initial_density is not consistent with inputs!"
                    + f" According to inputs, initial_density must be of shape {shape}, got {initial_density.shape}."
                )

            ics.hires_density = (
                Array(shape=shape, dtype=np.float32)
                .initialize()
                ._with_value_not_computed(val=initial_density)
            )
        return ics.compute()


@single_field_func
def perturb_field(
    *,
    redshift: float,
    inputs: InputParameters | None = None,
    initial_conditions: InitialConditions,
) -> PerturbedField:
    r"""
    Compute a perturbed field at a given redshift.

    Parameters
    ----------
    redshift : float
        The redshift at which to compute the perturbed field.
    initial_conditions : :class:`~InitialConditions` instance
        The initial conditions.

    Returns
    -------
    :class:`~PerturbedField`

    Other Parameters
    ----------------
    regenerate, write, cache:
        See docs of :func:`initial_conditions` for more information.

    Examples
    --------
    >>> initial_conditions = compute_initial_conditions()
    >>> field7 = perturb_field(7.0, initial_conditions)
    >>> field8 = perturb_field(8.0, initial_conditions)

    The user and cosmo parameter structures are by default inferred from the
    ``initial_conditions``.
    """
    # Initialize perturbed boxes.
    fields = PerturbedField.new(redshift=redshift, inputs=inputs)

    # Run the C Code
    return fields.compute(ics=initial_conditions)


@single_field_func
def determine_halo_catalog(
    *,
    redshift: float,
    inputs: InputParameters | None = None,
    initial_conditions: InitialConditions,
    descendant_halos: HaloCatalog | None = None,
) -> HaloCatalog:
    r"""
    Find a halo list, given a redshift.

    Parameters
    ----------
    redshift : float
        The redshift at which to determine the halo list.
    initial_conditions : :class:`~InitialConditions` instance
        The initial conditions fields (density, velocity).
    descendant_halos : :class:`~HaloCatalog` instance, optional
        The halos that form the descendants (i.e. lower redshift) of those computed by
        this function. If this is not provided, we generate the initial stochastic halos
        directly in this function (and progenitors can then be determined by these).

    Returns
    -------
    :class:`~HaloCatalog`

    Other Parameters
    ----------------
    regenerate, write, cache:
        See docs of :func:`initial_conditions` for more information.
    """
    if inputs.matter_options.HMF != "ST":
        warnings.warn(
            "DexM Halofinder sses a fit to the Sheth-Tormen mass function."
            "With HMF!=1 the Halos from DexM will not be from the same mass function",
            stacklevel=2,
        )

    if descendant_halos is None:
        descendant_halos = HaloCatalog.dummy()

    # Initialize halo list boxes.
    fields = HaloCatalog.new(
        redshift=redshift,
        desc_redshift=descendant_halos.redshift,
        inputs=inputs,
    )

    # Run the C Code
    return fields.compute(
        ics=initial_conditions,
        descendant_halos=descendant_halos,
    )


@single_field_func
def perturb_halo_catalog(
    *,
    initial_conditions: InitialConditions,
    inputs: InputParameters | None = None,
    previous_spin_temp: TsBox | None = None,
    previous_ionize_box: IonizedBox | None = None,
    halo_catalog: HaloCatalog,
) -> PerturbedHaloCatalog:
    r"""
    Given a halo list, perturb the halos for a given redshift.

    Parameters
    ----------
    initial_conditions : :class:`~InitialConditions`
        The initial conditions of the run. The user and cosmo params
        as well as the random seed will be set from this object.
    halo_catalog: :class: `~HaloCatalog`
        The halo catalogue in Lagrangian space to be perturbed.

    Returns
    -------
    :class:`~PerturbedHaloCatalog`

    Other Parameters
    ----------------
    regenerate, write, direc:
        See docs of :func:`initial_conditions` for more information.

    Examples
    --------
    Fill this in once finalised

    """
    hbuffer_size = (
        halo_catalog.n_halos if halo_catalog.n_halos else halo_catalog.buffer_size
    )
    redshift = halo_catalog.redshift

    # Initialize halo list boxes.
    fields = PerturbedHaloCatalog.new(
        redshift=redshift,
        buffer_size=hbuffer_size,
        inputs=inputs,
    )
    if previous_spin_temp is None:
        if (
            redshift >= inputs.simulation_options.Z_HEAT_MAX
            or not inputs.astro_options.USE_MINI_HALOS
        ):
            # Dummy spin temp is OK since we're above Z_HEAT_MAX
            previous_spin_temp = TsBox.dummy()
        else:
            raise ValueError("Below Z_HEAT_MAX you must specify the previous_spin_temp")

    if previous_ionize_box is None:
        if (
            redshift >= inputs.simulation_options.Z_HEAT_MAX
            or not inputs.astro_options.USE_MINI_HALOS
        ):
            # Dummy ionize box is OK since we're above Z_HEAT_MAX
            previous_ionize_box = IonizedBox.dummy()
        else:
            raise ValueError(
                "Below Z_HEAT_MAX you must specify the previous_ionize_box"
            )

    # Run the C Code
    return fields.compute(
        ics=initial_conditions,
        halo_catalog=halo_catalog,
        previous_spin_temp=previous_spin_temp,
        previous_ionize_box=previous_ionize_box,
    )


@single_field_func
def compute_halo_grid(
    *,
    redshift: float,
    initial_conditions: InitialConditions,
    inputs: InputParameters | None = None,
    halo_catalog: HaloCatalog | None = None,
    previous_spin_temp: TsBox | None = None,
    previous_ionize_box: IonizedBox | None = None,
) -> HaloBox:
    r"""
    Compute grids of halo properties from a catalogue.

    At the moment this simply produces halo masses, stellar masses and SFR on a grid of
    HII_DIM. In the future this will compute properties such as emissivities which will
    be passed directly into ionize_box etc. instead of the catalogue.

    Parameters
    ----------
    initial_conditions : :class:`~InitialConditions`
        The initial conditions of the run.
    inputs : :class:`~InputParameters`, optional
        The input parameters specifying the run.
    halo_catalog: :class:`~HaloCatalog`, optional
        This contains all the dark matter haloes obtained if using a discrete halo model.
        This is a list of halo masses and coordinates for the dark matter halos.
    previous_spin_temp : :class:`TsBox`, optional
        The previous spin temperature box. Used for feedback when USE_MINI_HALOS==True
    previous_ionize_box: :class:`IonizedBox` or None
        An at the last timestep. Used for feedback when USE_MINI_HALOS==True

    Returns
    -------
    :class:`~HaloBox` :
        An object containing the halo box data.

    Other Parameters
    ----------------
    regenerate, write, cache:
        See docs of :func:`initial_conditions` for more information.
    """
    box = HaloBox.new(redshift=redshift, inputs=inputs)

    if halo_catalog is None:
        if inputs.matter_options.has_discrete_halos:
            raise ValueError(
                f"You must provide halo_catalog for SOURCE_MODEL = {inputs.matter_options.SOURCE_MODEL}"
            )
        else:
            halo_catalog = HaloCatalog.dummy()

    # NOTE: due to the order, we use the previous spin temp here, like spin_temperature,
    #       but UNLIKE ionize_box, which uses the current box
    # TODO: think about the inconsistency here
    # NOTE: if USE_MINI_HALOS is TRUE, so is USE_TS_FLUCT and INHOMO_RECO
    if previous_spin_temp is None:
        if (
            redshift >= inputs.simulation_options.Z_HEAT_MAX
            or not inputs.astro_options.USE_MINI_HALOS
        ):
            # Dummy spin temp is OK since we're above Z_HEAT_MAX
            previous_spin_temp = TsBox.dummy()
        else:
            raise ValueError("Below Z_HEAT_MAX you must specify the previous_spin_temp")

    if previous_ionize_box is None:
        if (
            redshift >= inputs.simulation_options.Z_HEAT_MAX
            or not inputs.astro_options.USE_MINI_HALOS
        ):
            # Dummy ionize box is OK since we're above Z_HEAT_MAX
            previous_ionize_box = IonizedBox.dummy()
        else:
            raise ValueError(
                "Below Z_HEAT_MAX you must specify the previous_ionize_box"
            )

    return box.compute(
        initial_conditions=initial_conditions,
        halo_catalog=halo_catalog,
        previous_ionize_box=previous_ionize_box,
        previous_spin_temp=previous_spin_temp,
    )


# TODO: make this more general and probably combine with the lightcone interp function
def interp_halo_boxes(
    halo_boxes: list[HaloBox],
    fields: list[str],
    redshift: float,
) -> HaloBox:
    """
    Interpolate HaloBox history to the desired redshift.

    Photon conservation & Xray sources require halo boxes at redshifts
    that are not equal to the current redshift, and may be between redshift steps.
    So we need a function to interpolate between two halo boxes.
    We assume here that z_arr is strictly INCERASING

    Parameters
    ----------
    halo_boxes : list of HaloBox instances
        The halobox history to be interpolated
    fields: List of Strings
        The properties of the haloboxes to be interpolated
    redshift : float
        The desired redshift of interpolation

    Returns
    -------
    :class:`~HaloBox` :
        An object containing the halo box data
    """
    inputs = halo_boxes[0].inputs
    z_halos = [box.redshift for box in halo_boxes]
    if not np.all(np.diff(z_halos) > 0):
        raise ValueError("halo_boxes must be in ascending order of redshift")

    if redshift > z_halos[-1] or redshift < z_halos[0]:
        raise ValueError(f"Invalid z_target {redshift} for redshift array {z_halos}")

    # If we do global evolution, no need to do that
    if inputs.simulation_options.HII_DIM > 1:
        arr_fields = [f for f in fields if f in halo_boxes[0].arrays]
        computed = [box.ensure_arrays_computed(*arr_fields) for box in halo_boxes]
        if not all(computed):
            raise ValueError("Some of the HaloBox fields required are not computed")

    idx_prog = np.searchsorted(z_halos, redshift, side="left")

    if idx_prog == 0 or idx_prog == len(z_halos):
        logger.debug(f"redshift {redshift} beyond limits, {z_halos[0], z_halos[-1]}")
        raise ValueError

    z_prog = z_halos[idx_prog]
    idx_desc = idx_prog - 1

    z_desc = z_halos[idx_desc]
    interp_param = (redshift - z_desc) / (z_prog - z_desc)

    # If we do global evolution, no need to do that
    if inputs.simulation_options.HII_DIM > 1:
        # I set the box redshift to be the stored one so it is read properly into the ionize box
        # for the xray source it doesn't matter, also since it is not _compute()'d, it won't be cached
        check_output_consistency(
            dict(
                zip(
                    [f"box-{i}" for i in range(len(halo_boxes))],
                    halo_boxes,
                    strict=True,
                )
            )
        )
    hbox_out = HaloBox.new(redshift=redshift, inputs=inputs)

    # initialise the memory
    hbox_out._init_arrays()

    # interpolate halo boxes in gridded SFR
    hbox_prog = halo_boxes[idx_prog]
    hbox_desc = halo_boxes[idx_desc]

    for field in fields:
        field_desc = hbox_desc.get(field)
        field_prog = hbox_prog.get(field)
        interp_field = np.zeros_like(field_desc)
        interp_field[...] = (1 - interp_param) * field_desc + interp_param * field_prog
        hbox_out.set(field, interp_field)

    return hbox_out


# NOTE: the current implementation of this box is very hacky, since I have trouble figuring out a way to _compute()
#   over multiple redshifts in a nice way using this wrapper.
# TODO: if we move some code to jax or similar I think this would be one of the first candidates (just filling out some filtered grids)
@single_field_func
def compute_xray_source_field(
    *,
    initial_conditions: InitialConditions,
    hboxes: list[HaloBox],
    redshift: float,
) -> XraySourceBox:
    r"""
    Compute filtered grid of SFR for use in spin temperature calculation.

    This will filter over the halo history in annuli, computing the contribution to the
    SFR density

    If no halo field is passed one is calculated at the desired redshift as if it is the
    first box.

    Parameters
    ----------
    initial_conditions : :class:`~InitialConditions`
        The initial conditions of the run. The user and cosmo params
    hboxes: Sequence of :class:`~HaloBox` instances
        This contains the list of Halobox instances which are used to create this source field

    Returns
    -------
    :class:`~XraySourceBox` :
        An object containing x ray heating, ionisation, and lyman alpha rates.

    Other Parameters
    ----------------
    regenerate, write, cache:
        See docs of :func:`initial_conditions` for more information.
    """
    z_halos = [hb.redshift for hb in hboxes]
    inputs = hboxes[0].inputs

    # Initialize halo list boxes.
    box = XraySourceBox.new(redshift=redshift, inputs=inputs)

    # set minimum R at cell size
    l_factor = (4 * np.pi / 3.0) ** (-1 / 3)
    if inputs.simulation_options.HII_DIM == 1:
        # If HII_DIM=1 (happens when we run_global_evolution), we take a typical cell size of 1.5Mpc,
        # just to for setting the z'' array (note that filtering won't be done on a box with a single cell)
        R_min = 1.5 * l_factor
    else:
        R_min = (
            inputs.simulation_options.BOX_LEN
            / inputs.simulation_options.HII_DIM
            * l_factor
        )
    z_max = min(max(z_halos), inputs.simulation_options.Z_HEAT_MAX)

    # now we need to find the closest halo box to the redshift of the shell
    cosmo_ap = inputs.cosmo_params.cosmo
    cmd_zp = cosmo_ap.comoving_distance(redshift)
    R_steps = np.arange(0, inputs.astro_params.N_STEP_TS)
    R_factor = (inputs.astro_params.R_MAX_TS / R_min) ** (
        R_steps / inputs.astro_params.N_STEP_TS
    )
    R_range = un.Mpc * R_min * R_factor
    cmd_edges = cmd_zp + R_range  # comoving distance edges
    # Get the edges of the shells
    zmin = z_at_value(cosmo_ap.comoving_distance, cmd_edges.min()).value
    zmax = z_at_value(cosmo_ap.comoving_distance, cmd_edges.max()).value
    zgrid = np.logspace(np.log10(zmin), np.log10(zmax), 100)
    dgrid = cosmo_ap.comoving_distance(zgrid)
    zpp_edges = np.interp(cmd_edges.value, dgrid.value, zgrid)
    # the `average` redshift of the shell is the average of the
    # inner and outer redshifts (following the C code)
    zpp_avg = zpp_edges - np.diff(np.insert(zpp_edges, 0, redshift)) / 2

    interp_fields = ["halo_sfr", "halo_xray"]
    if inputs.astro_options.USE_MINI_HALOS:
        interp_fields += ["halo_sfr_mini", "log10_Mcrit_MCG_ave"]

    # call the box the initialize the memory, since I give some values before computing
    box._init_arrays()
    for i in range(inputs.astro_params.N_STEP_TS):
        R_inner = R_range[i - 1].to("Mpc").value if i > 0 else 0
        R_outer = R_range[i].to("Mpc").value

        if zpp_avg[i] >= z_max:
            box.filtered_sfr.value[i] = 0
            box.filtered_xray.value[i] = 0
            if inputs.astro_options.USE_MINI_HALOS:
                box.filtered_sfr_mini.value[i] = 0
                box.mean_log10_Mcrit_LW.value[i] = inputs.astro_params.M_TURN  # minimum
            logger.debug(f"ignoring Radius {i} which is above Z_HEAT_MAX")
            continue

        hbox_interp = interp_halo_boxes(
            halo_boxes=hboxes[::-1],
            fields=interp_fields,
            redshift=zpp_avg[i],
        )

        # if we have no halos we ignore the whole shell
        sfr_allzero = np.all(hbox_interp.get("halo_sfr") == 0)
        if inputs.astro_options.USE_MINI_HALOS:
            sfr_allzero = sfr_allzero & np.all(hbox_interp.get("halo_sfr_mini") == 0)
        if sfr_allzero:
            box.filtered_sfr.value[i] = 0
            box.filtered_xray.value[i] = 0
            if inputs.astro_options.USE_MINI_HALOS:
                box.filtered_sfr_mini.value[i] = 0
                box.mean_log10_Mcrit_LW.value[i] = hbox_interp.log10_Mcrit_MCG_ave
            logger.debug(f"ignoring Radius {i} due to no stars")
            continue

        box = box.compute(
            halobox=hbox_interp,
            R_inner=R_inner,
            R_outer=R_outer,
            R_ct=i,
            allow_already_computed=True,
        )

    # Sometimes we don't compute at all
    # (if the first zpp > z_max or there are no halos at max R)
    # in which case the array is not marked as computed
    if not box.is_computed:
        for name, array in box.arrays.items():
            setattr(box, name, array.computed())

    return box


@single_field_func
def compute_spin_temperature(
    *,
    initial_conditions: InitialConditions,
    perturbed_field: PerturbedField,
    inputs: InputParameters | None = None,
    xray_source_box: XraySourceBox | None = None,
    previous_spin_temp: TsBox | None = None,
    cleanup: bool = False,
) -> TsBox:
    r"""
    Compute spin temperature boxes at a given redshift.

    See the notes below for how the spin temperature field is evolved through redshift.

    Parameters
    ----------
    initial_conditions : :class:`~InitialConditions`
        The initial conditions
    inputs : :class:`~InputParameters`
        The input parameters specifying the run. Since this will be the first box
        to use the astro params/flags when SOURCE_MODEL='E-INTEGRAL' and USE_TS_FLUCT=True.
    perturbed_field : :class:`~PerturbedField`, optional
        If given, this field will be used, otherwise it will be generated. To be generated,
        either `initial_conditions` and `redshift` must be given, or `simulation_options`, `cosmo_params` and
        `redshift`. By default, this will be generated at the same redshift as the spin temperature
        box. The redshift of perturb field is allowed to be different than `redshift`. If so, it
        will be interpolated to the correct redshift, which can provide a speedup compared to
        actually computing it at the desired redshift.
    xray_source_box : :class:`XraySourceBox`, optional
        When using a lagrangian source model, this box specifies the filtered sfr and xray emissivity at all
        redshifts/filter radii required by the spin temperature algorithm.
    previous_spin_temp : :class:`TsBox` or None
        The previous spin temperature box. Needed when we are beyond the first snapshot

    Returns
    -------
    :class:`~TsBox`
        An object containing the spin temperature box data.

    Other Parameters
    ----------------
    regenerate, write, cache:
        See docs of :func:`initial_conditions` for more information.
    """
    redshift = perturbed_field.redshift

    if redshift >= inputs.simulation_options.Z_HEAT_MAX:
        previous_spin_temp = TsBox.new(inputs=inputs, redshift=0.0, dummy=True)

    if xray_source_box is None:
        if inputs.matter_options.lagrangian_source_grid:
            raise ValueError(
                f"xray_source_box is required for SOURCE_MODEL= {inputs.matter_options.SOURCE_MODEL}"
            )
        else:
            xray_source_box = XraySourceBox.dummy()

    # Set up the box without computing anything.
    box = TsBox.new(
        redshift=redshift,
        inputs=inputs,
    )

    # Run the C Code
    return box.compute(
        cleanup=cleanup,
        perturbed_field=perturbed_field,
        xray_source_box=xray_source_box,
        prev_spin_temp=previous_spin_temp,
        ics=initial_conditions,
    )


@single_field_func
def compute_ionization_field(
    *,
    perturbed_field: PerturbedField,
    initial_conditions: InitialConditions,
    inputs: InputParameters | None = None,
    previous_perturbed_field: PerturbedField | None = None,
    previous_ionized_box: IonizedBox | None = None,
    spin_temp: TsBox | None = None,
    halobox: HaloBox | None = None,
) -> IonizedBox:
    r"""
    Compute an ionized box at a given redshift.

    This function has various options for how the evolution of the ionization is
    computed (if at all). See the Notes below for details.

    Parameters
    ----------
    initial_conditions : :class:`~InitialConditions` instance
        The initial conditions.
    inputs : :class:`~InputParameters`
        The input parameters specifying the run. Since this may be the first box
        to use the astro params/flags, it is needed when we have not computed a TsBox or HaloBox.
    perturbed_field : :class:`~PerturbedField`
        The perturbed density field.
    previous_perturbed_field : :class:`~PerturbedField`, optional
        An perturbed field at higher redshift. This is only used if USE_MINI_HALOS is included.
    previous_ionize_box: :class:`IonizedBox` or None
        An ionized box at higher redshift. This is only used if `INHOMO_RECO` and/or `USE_TS_FLUCT`
        are true. If either of these are true, and this is not given, then it will be assumed that
        this is the "first box", i.e. that it can be populated accurately without knowing source
        statistics.
    spin_temp: :class:`TsBox` or None, optional
        A spin-temperature box, only required if `USE_TS_FLUCT` is True. If None, will try to read
        in a spin temp box at the current redshift, and failing that will try to automatically
        create one, using the previous ionized box redshift as the previous spin temperature
        redshift.
    halobox: :class:`~HaloBox` or None, optional
        If passed, this contains all the dark matter haloes obtained if using the
        lagrangian source models. These are grids containing summed halo properties
        such as ionizing emissivity.

    Returns
    -------
    :class:`~IonizedBox` :
        An object containing the ionized box data.

    Notes
    -----
    Typically, the ionization field at any redshift is dependent on the evolution of xHI up until
    that redshift, which necessitates providing a previous ionization field to define the current
    one. If neither the spin temperature field, nor inhomogeneous recombinations (specified in
    flag options) are used, no evolution needs to be done. If the redshift is beyond
    Z_HEAT_MAX, previous fields are not required either.
    """
    redshift = perturbed_field.redshift

    if redshift >= inputs.simulation_options.Z_HEAT_MAX:
        # Previous boxes must be "initial"
        previous_ionized_box = IonizedBox.initial(inputs=inputs)
        previous_perturbed_field = PerturbedField.initial(inputs=inputs)

    if inputs.evolution_required:
        if previous_ionized_box is None:
            raise ValueError(
                "You need to provide a previous ionized box when redshift < Z_HEAT_MAX."
            )
        if previous_perturbed_field is None:
            raise ValueError(
                "You need to provide a previous perturbed field when redshift < Z_HEAT_MAX."
            )
    else:
        if previous_ionized_box is None:
            previous_ionized_box = IonizedBox.initial(inputs=inputs)
        if previous_perturbed_field is None:
            previous_perturbed_field = PerturbedField.initial(inputs=inputs)

    if inputs.simulation_options.HII_DIM > 1:
        box = IonizedBox.new(inputs=inputs, redshift=redshift)

        if not inputs.matter_options.lagrangian_source_grid:
            # Construct an empty halo field to pass in to the function.
            halobox = HaloBox.dummy()
        elif halobox is None:
            raise ValueError(
                f"A HaloBox must be provided for SOURCE_MODEL={inputs.matter_options.SOURCE_MODEL}"
            )

        # Set empty spin temp box if necessary.
        if not inputs.astro_options.USE_TS_FLUCT:
            spin_temp = TsBox.dummy()
        elif spin_temp is None:
            raise ValueError("No spin temperature box given but USE_TS_FLUCT=True")

        # Run the C Code
        return box.compute(
            perturbed_field=perturbed_field,
            prev_perturbed_field=previous_perturbed_field,
            prev_ionize_box=previous_ionized_box,
            spin_temp=spin_temp,
            halobox=halobox,
            ics=initial_conditions,
        )
    else:
        # If we have only one cell (could happen if we do a global evolution), we set the neutral fraction
        # according to the global evolution

        from .global_evolution import (
            compute_global_reionization_at_z,  # This is imported here to prevent circular import
        )

        box = compute_global_reionization_at_z(
            redshift=redshift,
            inputs=inputs,
            previous_ionized_box=previous_ionized_box,
            spin_temp=spin_temp,
        )
        return box


@single_field_func
def brightness_temperature(
    *,
    ionized_box: IonizedBox,
    perturbed_field: PerturbedField,
    spin_temp: TsBox | None = None,
) -> BrightnessTemp:
    r"""
    Compute a coeval brightness temperature box.

    Parameters
    ----------
    ionized_box: :class:`IonizedBox`
        A pre-computed ionized box.
    perturbed_field: :class:`PerturbedField`
        A pre-computed perturbed field at the same redshift as `ionized_box`.
    spin_temp: :class:`TsBox`, optional
        A pre-computed spin temperature, at the same redshift as the other boxes.

    Returns
    -------
    :class:`BrightnessTemp` instance.
    """
    redshift = ionized_box.redshift
    inputs = ionized_box.inputs

    if spin_temp is None:
        if inputs.astro_options.USE_TS_FLUCT:
            raise ValueError(
                "You have USE_TS_FLUCT=True, but have not provided a spin_temp!"
            )
        else:
            spin_temp = TsBox.dummy()

    box = BrightnessTemp.new(redshift=redshift, inputs=inputs)

    return box.compute(
        spin_temp=spin_temp,
        ionized_box=ionized_box,
        perturbed_field=perturbed_field,
    )



"""
Unit-tests of the coeval driver.

They do not test for correctness of simulations, but whether different parameter options
work/don't work as intended.
"""

import attrs
import pytest

import py21cmfast as p21c
from py21cmfast import Coeval, run_coeval
from py21cmfast.wrapper.arrays import Array
from py21cmfast.wrapper.outputs import OutputStruct


def test_coeval_st(ic, default_input_struct_ts, cache):
    coeval = run_coeval(
        initial_conditions=ic,
        inputs=default_input_struct_ts,
        cache=cache,
    )
    assert isinstance(coeval[0].ts_box, p21c.TsBox)


def test_run_coeval_bad_inputs(ic, perturbed_field, default_input_struct, cache):
    with pytest.raises(
        ValueError, match="out_redshifts must be given if inputs has no node redshifts"
    ):
        run_coeval(initial_conditions=ic, inputs=default_input_struct, cache=cache)


def test_coeval_lowerz_than_photon_cons(
    ic, default_input_struct, default_astro_options, cache
):
    with pytest.raises(ValueError, match="You have passed a redshift"):
        run_coeval(
            initial_conditions=ic,
            out_redshifts=2.0,
            inputs=default_input_struct.clone(
                astro_options=default_astro_options.clone(
                    PHOTON_CONS_TYPE="z-photoncons",
                )
            ),
            cache=cache,
        )


def test_coeval_warnings(default_input_struct_lc, cache):
    # test for no caching with halo fields
    inputs = default_input_struct_lc.evolve_input_structs(
        SOURCE_MODEL="CHMF-SAMPLER",
    )

    with pytest.warns(UserWarning, match="You have turned off caching"):
        run_coeval(
            out_redshifts=16.0,
            inputs=inputs,
            write=False,
            cache=cache,
        )

    inputs = default_input_struct_lc.evolve_input_structs(
        USE_TS_FLUCT=True,
    )

    # test for minimum node redshift > out_redshifts
    with pytest.warns(UserWarning, match="minimum node redshift"):
        run_coeval(
            out_redshifts=8.0,
            inputs=inputs,
            write=False,
            cache=cache,
        )


def test_coeval_fields():
    """Test that every array ends up in the coeval object."""
    fields = Coeval.get_fields()
    for kls in OutputStruct.__subclasses__():
        for field in attrs.fields(kls):
            if isinstance(field, Array):
                assert field.name in fields



"""Utilities that help with wrapping various C structures."""

import logging

import numpy as np
from cffi import FFI

from ..c_21cmfast import ffi, lib
from .exceptions import _process_exitcode

_ffi = FFI()

logger = logging.getLogger(__name__)


ctype2dtype = {}

# Integer types
for prefix in ("int", "uint"):
    for log_bytes in range(4):
        ctype = f"{prefix}{8 * 2**log_bytes:d}_t"
        dtype = f"{prefix[0]}{2**log_bytes:d}"
        ctype2dtype[ctype] = np.dtype(dtype)

# Floating point types
ctype2dtype["float"] = np.dtype("f4")
ctype2dtype["double"] = np.dtype("f8")
ctype2dtype["int"] = np.dtype("i4")


def asarray(ptr, shape):
    """Get the canonical C type of the elements of ptr as a string."""
    ctype = _ffi.getctype(_ffi.typeof(ptr).item).split("*")[0].strip()

    if ctype not in ctype2dtype:
        raise RuntimeError(
            f"Cannot create an array for element type: {ctype}. Can do {list(ctype2dtype.values())}."
        )

    array = np.frombuffer(
        _ffi.buffer(ptr, _ffi.sizeof(ctype) * np.prod(shape)), ctype2dtype[ctype]
    )
    array.shape = shape
    return array


def _call_c_simple(fnc, *args):
    """Call a simple C function that just returns an object.

    Any such function should be defined such that the last argument is an int pointer generating
    the status.
    """
    # Parse the function to get the type of the last argument
    cdata = str(ffi.addressof(lib, fnc.__name__))
    kind = cdata.split("(")[-1].split(")")[0].split(",")[-1]
    result = ffi.new(kind)
    status = fnc(*args, result)
    _process_exitcode(status, fnc, args)
    return result[0]


def camel_to_snake(word: str, depublicize: bool = False):
    """Convert came case to snake case."""
    word = "".join(f"_{i.lower()}" if i.isupper() else i for i in word)

    if not depublicize:
        word = word.lstrip("_")

    return word


def snake_to_camel(word: str, publicize: bool = True):
    """Convert snake case to camel case."""
    if "_" not in word:
        return word

    if publicize:
        word = word.lstrip("_")
    return "".join(x.capitalize() or "_" for x in word.split("_"))


def get_all_subclasses(cls):
    """Get a list of all subclasses of a given class, recursively."""
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


def float_to_string_precision(x, n):
    """Print out a standard float number at a given number of significant digits.

    Code here: https://stackoverflow.com/a/48812729
    """
    return f"{float(f'{x:.{int(n)}g}'):g}"



"""Test CLI functionality."""

import tempfile
import tomllib as toml
from pathlib import Path

import pytest
from rich.console import Console

from py21cmfast import Coeval, LightCone, cli
from py21cmfast._templates import create_params_from_template
from py21cmfast.cli import Parameters, ParameterSelection, RunParams, _run_setup, app
from py21cmfast.io.h5 import read_output_struct


def app_noexit(*args, **kwargs):
    """Return the CLI app with SystemExit disabled for testing."""
    return app(*args, **kwargs, result_action="return_value")


class TestTemplateAvail:
    """Tests of the template avail command."""

    def test_that_it_prints(self, capsys):
        """Test that it prints out and contains known template names."""
        app_noexit("template avail")
        output = capsys.readouterr().out
        assert "simple" in output
        assert "Munoz21" in output


class TestTemplateCreate:
    """Test the `template create` command."""

    def test_create_without_explicit_params(self, tmp_path: Path):
        """Test creating from a template without overriding doesn't change anything."""
        app_noexit(
            f"template create --template simple --out {tmp_path / 'simple.toml'}"
        )
        assert (tmp_path / "simple.toml").exists()

        p1 = create_params_from_template(tmp_path / "simple.toml")
        p2 = create_params_from_template("simple")

        assert all(v == p2[k] for k, v in p1.items())

    def test_create_with_explicit_params(self, tmp_path: Path):
        """Test that overriding params does change the inputs."""
        out = tmp_path / "simple_plus.toml"
        app_noexit(f"template create --template simple --hii-dim 37 --out {out}")
        assert out.exists()

        p1 = create_params_from_template(out)
        p2 = create_params_from_template("simple")

        assert p1["simulation_options"].HII_DIM == 37
        assert p2["simulation_options"].HII_DIM != 37

    def test_failure_with_both_template_and_file(self, tmp_path):
        """Test that providing both --template and --param-file errors."""
        new = tmp_path / "new.toml"
        app_noexit(f"template create --template simple --out {new}")

        # This should fail
        with pytest.raises(SystemExit):
            app_noexit(
                f"template create --param-file {new} --template simple --out here.toml"
            )

    def test_default_minimal(self, tmp_path: Path):
        """Test that creating a minimal toml with no values works."""
        fl = tmp_path / "test.toml"
        app_noexit(f"template create --mode minimal --out {fl}")

        with fl.open("rb") as _fl:
            data = toml.load(_fl)

        # Since it was minimal, the file should have no values.
        assert all(len(v) == 0 for v in data.values())

    def test_non_existent_directory(self, tmp_path: Path):
        """Test that providing a non-existing directory to write to is OK."""
        out = tmp_path / "parent" / "config.toml"
        app_noexit(f"template create --template simple --out {out}")
        assert out.exists()


class TestTemplateShow:
    """Tests of the `template show` command."""

    def test_show_alias(self, capsys):
        """Test that showing an alias works."""
        app_noexit("template show EOS21")
        output = capsys.readouterr().out
        assert "from Munoz" in output


class TestRunSetup:
    """Tests of the _run_setup function."""

    def setup_class(self):
        """Make a default temp dir, and store a simple config TOML."""
        self.tmpdir = Path(tempfile.mkdtemp())

        self.simple = self.tmpdir / "simple.toml"

        # Create a full template in tmpdir
        app_noexit(f"template create --template simple --out {self.simple}")

    def test_unmodified_paramfile(self, capsys):
        """Test that running with an unmodified --param-file doesn't write the fullspec."""
        runp = RunParams(
            param_selection=ParameterSelection(param_file=[self.simple]),
            cachedir=self.tmpdir,
        )
        params = Parameters()  # don't modify the input
        _run_setup(runp, params)
        out = capsys.readouterr().out
        assert "Wrote full configuration" not in out

    def test_unmodified_template(self, capsys):
        """Test that an unmodified --template does write a simple fullspec TOML."""
        runp = RunParams(
            param_selection=ParameterSelection(template=["simple"]),
            cachedir=self.tmpdir,
        )
        params = Parameters()  # don't modify the input
        _run_setup(runp, params)
        out = capsys.readouterr().out
        assert "Wrote full configuration" in out
        assert "simple.toml" in out

    def test_explicit_outcfg(self, capsys):
        """Test that directly modifying params and passing an explicit file works."""
        outcfg = self.tmpdir / "custom-name.toml"
        runp = RunParams(
            param_selection=ParameterSelection(template=["simple"]),
            outcfg=outcfg,
            cachedir=self.tmpdir,
        )
        params = Parameters(simulation_options=cli._SimulationOptions(HII_DIM=37))

        _run_setup(runp, params)
        out = capsys.readouterr().out
        assert "Wrote full configuration" in out
        assert f"{outcfg}" in out

        _run_setup(runp, Parameters())
        out = capsys.readouterr().out
        assert "Wrote full configuration" in out
        assert f"{outcfg}" in out

    def test_unknown_name(self, capsys):
        """Test that modifying params without an explicit file creates a random file."""
        runp = RunParams(
            param_selection=ParameterSelection(template=["simple"]),
            cachedir=self.tmpdir,
        )
        params = Parameters(simulation_options=cli._SimulationOptions(HII_DIM=37))

        _run_setup(runp, params)
        out = capsys.readouterr().out
        assert "Wrote full configuration" in out
        assert "simple.toml" not in out  # got a random uuid


class TestRunICS:
    """Tests of the `run ics` command."""

    def test_basic_run(self, capsys, tmp_path: Path):
        """Test that a simple run creates an InitialConditions.h5 file."""
        app_noexit(
            f"run ics --template simple tiny --cachedir {tmp_path}",
            console=Console(width=100),
        )
        output = capsys.readouterr().out
        assert "Saved initial conditions" in output

        outfile = Path(output.split("conditions to ")[-1].replace("\n", ""))
        assert outfile.exists()
        ics = read_output_struct(outfile)
        assert ics.simulation_options.HII_DIM == 32

    def test_warn_formatting(self, tmp_path, capsys):
        """Test that warnings are printed properly."""
        app_noexit(
            f"run ics --template simple tiny --box-len 400 --zmin 5.0 --cachedir {tmp_path}"
        )
        out = capsys.readouterr().out
        assert "Resolution is likely too low" in out

    def test_regen(self, capsys, tmp_path):
        """Test that re-running the same box with --regen does actually re-run things."""
        app_noexit(f"run ics --template simple tiny --cachedir {tmp_path}")

        # Now run it again right away with regen
        app_noexit(
            f"run ics --template simple tiny --cachedir {tmp_path} --regenerate",
        )
        out = capsys.readouterr().out
        assert "regeneration is requested. Overriding." in out

        # Run it without regen
        app_noexit(f"run ics --template simple tiny --cachedir {tmp_path}")
        out = capsys.readouterr().out
        assert "skipping computation" in out


class TestRunCoeval:
    """Tests of the `run coeval` command."""

    def test_basic_run(self, capsys, tmp_path: Path):
        """Test that a basic run through produces a coeval*.h5 file."""
        cfile = tmp_path / "coeval_z6.00.h5"
        app_noexit(
            f"run coeval --template simple tiny --cachedir {tmp_path} "
            f"--redshifts 6.0 --out {cfile.parent}",
        )

        output = capsys.readouterr().out
        assert "Saved z=6.00 coeval box" in output

        assert cfile.exists()
        cv = Coeval.from_file(cfile)
        assert cv.redshift == 6.0

    def test_node_redshifts(self, capsys, tmp_path):
        """Test that having nodez in addition to --redshifts works."""
        # We have other node redshifts, but we don't do anything with them.
        app_noexit(
            f"run coeval --template Park19 tiny --zprime-step-factor 1.4 --z-heat-max 15 "
            f"--cachedir {tmp_path} "
            f"--no-save-all-redshifts "
            f"--redshifts 6.0 --out {tmp_path}",
        )
        cfile = tmp_path / "coeval_z6.00.h5"
        assert cfile.exists()

        # This time save everything....
        new = tmp_path / "new"
        new.mkdir()
        app_noexit(
            f"run coeval --template Park19 tiny --cachedir {new} "
            f"--save-all-redshifts "
            f"--redshifts 6.0 --out {new}",
        )
        assert len(list(new.glob("coeval*.h5"))) > 1


class TestRunLightcone:
    """Test the `run lightcone` command."""

    def test_basic_run(self, capsys, tmp_path: Path):
        """Test that a basic run produces a lightcone.h5 file."""
        lcfile = tmp_path / "lightcone.h5"
        app_noexit(
            f"run lightcone --template simple tiny --cachedir {tmp_path} "
            f"--redshift-range 6.0 12.0 --out {lcfile}",
        )

        output = capsys.readouterr().out
        assert "Saved Lightcone" in output

        assert lcfile.exists()
        LightCone.from_file(lcfile)

    def test_non_existent_path(self, tmp_path):
        """Test that a non-existent output path is OK."""
        lcfile = tmp_path / "new" / "lightcone.h5"
        app_noexit(
            f"run lightcone --template simple tiny --cachedir {tmp_path} "
            f"--redshift-range 6.0 12.0 --out {lcfile}",
        )

        assert lcfile.exists()


class TestParamHelp:
    """Test the `run params` command."""

    def test_printing(self, capsys):
        """Test that the (stub) command prints a short useful message."""
        app_noexit("run params")
        assert "Usage: 21cmfast run params --help" in capsys.readouterr().out

    def test_full_help(self, capsys):
        """Test that the --help command prints out all the param help."""
        app_noexit("run params --help")
        out = capsys.readouterr().out

        assert "--hii-dim" in out
        assert "SimulationOptions" in out
        assert "--use-ts-fluct" in out


class TestPRFeature:
    """Test the `dev feature` command."""

    def test_simple_run_through(self, tmp_path: Path):
        """Test that a simple run-through produces the expected plots."""
        template = tmp_path / "small-simple.toml"
        app_noexit(f"template create --template simple tiny --out {template}")
        app_noexit(
            f"dev feature --param-file {template} --redshift-range 6 12 --hmf PS --cachedir {tmp_path} --outdir {tmp_path}"
        )
        assert (tmp_path / "pr_feature_history.pdf").exists()
        assert (tmp_path / "pr_feature_power_history.pdf").exists()
        assert (tmp_path / "pr_feature_lightcone_2d_brightness_temp.pdf").exists()


class TestPredictStructSize:
    """Test the predict struct-size command."""

    def test_relevant_text_is_printed(self, capsys):
        """Test that running the size prediction CLI prints relevant text."""
        app_noexit("predict struct-size --template simple tiny --unit mb")
        # We just want to make sure it runs and prints something reasonable.
        # Detailed correctness is tested in management.py tests.

        out = capsys.readouterr().out
        assert "Output Struct Sizes" in out
        assert "InitialConditions" in out

    def test_cache_off(self, capsys):
        """Test that running with cache off affects the predicted sizes."""
        app_noexit("predict struct-size --template simple tiny --cache-config off")
        out_off = capsys.readouterr().out

        app_noexit("predict struct-size --template simple tiny --cache-config on")
        out_on = capsys.readouterr().out

        assert out_off != out_on
        assert "InitialConditions" in out_on
        assert "InitialConditions" not in out_off


class TestPredictTotalStorageSize:
    """Test the predict total storage-size command."""

    @pytest.mark.parametrize(
        "template",
        ["simple tiny", "Park19 small", "Munoz21 small", "latest-dhalos large"],
    )
    def test_relevant_text_is_printed(self, capsys, template: str):
        """Test that running the total storage size CLI prints relevant text."""
        app_noexit(f"predict storage-size --template {template} --unit gb")
        # We just want to make sure it runs and prints something reasonable.
        # Detailed correctness is tested in management.py tests.

        out = capsys.readouterr().out
        assert "Storage Sizes" in out

    @pytest.mark.parametrize(
        "template",
        ["simple tiny", "Park19 small", "Munoz21 small", "latest-dhalos large"],
    )
    def test_cache_off(self, capsys, template: str):
        """Test that running with cache off affects the predicted total storage size."""
        app_noexit(f"predict storage-size --template {template} --cache-config off")
        out_off = capsys.readouterr().out

        app_noexit("predict storage-size --template simple tiny --cache-config on")
        out_on = capsys.readouterr().out

        assert out_off != out_on
        assert "PerturbedField" in out_on
        assert "PerturbedField" not in out_off



"""Build the C code with CFFI."""

import os
import sys
from pathlib import Path

from cffi import FFI


# Get the compiler. We support gcc and clang.
# The compiler is determnined from the environment and uses sysconfig as a fallback.
def detect_default_compiler():
    """Detect the default compiler using its "--version" command.

    CFFI gives us no control over which compiler is used, so we have to
    detect which compiler is being used based on the environment or sysconfig.

    Additionally, it uses the deprecated distutils package to get the default compiler,
    so we avoid that here. Instead calling the compiler with "--version" and looking for
    substrings.

    Returns a string indicating the compiler type: "gcc", "clang", or "unknown".
    This allows us to set compiler-specific flags later on.
    """
    import sysconfig
    import warnings
    from subprocess import run

    source = "environment variable 'CC'" if "CC" in os.environ else "sysconfig"
    _compiler = os.environ.get("CC", sysconfig.get_config_var("CC"))

    # strip any arguments from the compiler command
    _compiler = _compiler.split()[0]

    print(f"Using compiler from {source}: {_compiler}")

    result = run([_compiler, "--version"], capture_output=True, text=True)
    if any(s in result.stdout for s in ["gcc", "g++", "gnu"]):
        return "gcc"
    elif "clang" in _compiler.lower():
        return "clang"
    else:
        warnings.warn(
            f"21cmFAST was not able to determine the compiler type from '{_compiler}'."
            " No special compiler flags will be set.",
            stacklevel=2,
        )
    return "unknown"


ffi = FFI()

LOCATION = Path(__file__).resolve().parent
CLOC = LOCATION / "src" / "py21cmfast" / "src"
include_dirs = [str(CLOC)]
c_files = [str(fl.relative_to(LOCATION)) for fl in sorted(CLOC.glob("*.c"))]

# Set the C-code logging level.
# If DEBUG is set, we default to the highest level, but if not,
# we set it to the level just above no logging at all.
log_level = os.environ.get("LOG_LEVEL", 4 if "DEBUG" in os.environ else 1)
available_levels = [
    "NONE",
    "ERROR",
    "WARNING",
    "INFO",
    "DEBUG",
    "SUPER_DEBUG",
    "ULTRA_DEBUG",
]


if isinstance(log_level, str) and log_level.upper() in available_levels:
    log_level = available_levels.index(log_level.upper())

try:
    log_level = int(log_level)
except ValueError as e:
    # note: for py35 support, can't use f strings.
    raise ValueError(
        "LOG_LEVEL must be specified as a positive integer, or one "
        f"of {available_levels}"
    ) from e

# ==================================================
# Set compilation arguments dependent on environment
# ==================================================

extra_compile_args = ["-Wall", "--verbose", f"-DLOG_LEVEL={log_level:d}"]

if "DEBUG" in os.environ:
    extra_compile_args += ["-g", "-O0"]
else:
    extra_compile_args += ["-Ofast"]

if sys.platform == "darwin":
    extra_compile_args += ["-Xpreprocessor"]

extra_compile_args += ["-fopenmp"]

libraries = ["m", "gsl", "gslcblas", "fftw3f_omp", "fftw3f"]

# stuff for gperftools
if "PROFILE" in os.environ:
    libraries += ["profiler", "tcmalloc"]
    # we need this even if DEBUG is off
    extra_compile_args += ["-g"]

# do compiler-specific stuff here
compiler = detect_default_compiler()
if compiler == "clang":
    libraries += ["omp"]

library_dirs = []
for k, v in os.environ.items():
    if "inc" in k.lower():
        include_dirs += [v]
    elif "lib" in k.lower():
        library_dirs += [v]

# =================================================================
# NOTES FOR DEVELOPERS:
#   The CFFI implementation works as follows:
#       - All function prototypes, global variables and type definitions *directly* used
#           in the python wrapper must be declared via ffi.cdef("""C CODE""").
#           There must be no compiler directives in this code (#include, #define, etc)
#       - All implementations of global variables and types present in the cdef() calls
#           must also be present in the second argument of set_source.
#           This is passed to the compiler.
#       - The `sources` kwarg then contains all the .c files in the library which are to be compiled

# This is the overall C code.
ffi.set_source(
    "py21cmfast.c_21cmfast",  # Name/Location of shared library module
    """
    #include "21cmFAST.h"
    """,
    sources=c_files,
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    libraries=libraries,
    extra_compile_args=extra_compile_args,
)

# Header files containing types, globals and function prototypes
with (CLOC / "_inputparams_wrapper.h").open() as f:
    ffi.cdef(f.read())
with (CLOC / "_outputstructs_wrapper.h").open() as f:
    ffi.cdef(f.read())
with (CLOC / "_functionprototypes_wrapper.h").open() as f:
    ffi.cdef(f.read())

# CFFI needs to be able to access a free function to make the __del__ method for OutputStruct fields
#  This will expose the standard free() function to the wrapper so it can be used
ffi.cdef(
    """
        void free(void *ptr);
    """
)

if __name__ == "__main__":
    ffi.compile()



"""Tests of the halo sampler."""

import matplotlib as mpl
import numpy as np
import pytest

from py21cmfast import (
    IonizedBox,
    TsBox,
    compute_initial_conditions,
    determine_halo_catalog,
    perturb_halo_catalog,
)
from py21cmfast.wrapper import cfuncs as cf

from . import test_c_interpolation_tables as cint
from .produce_integration_test_data import get_all_options_struct, print_failure_stats

RELATIVE_TOLERANCE = 1e-1

options_hmf = list(cint.OPTIONS_HMF.keys())
options_delta = [-0.9, -0.5, 0, 1, 1.4]  # cell densities to draw samples from
options_mass = [1e9, 1e10, 1e11, 1e12, 1e13]  # halo masses to draw samples from


@pytest.mark.parametrize("name", options_hmf)
@pytest.mark.parametrize("cond_type", ["cat", "grid"])
@pytest.mark.parametrize("cond", range(len(options_delta)))
def test_sampler(name, cond, cond_type, plt):
    _, kwargs = cint.OPTIONS_HMF[name]
    # we need the largest condition mass to not trip the extreme halo check
    # at the test redshift, since that test would not be very useful
    redshift = 7.0
    inputs = get_all_options_struct(redshift, **kwargs)["inputs"]
    inputs = inputs.evolve_input_structs(
        SAMPLER_MIN_MASS=min(options_mass) / 2,
    )

    from_cat = "cat" in cond_type
    z_desc = (
        (redshift + 1) / inputs.simulation_options.ZPRIME_STEP_FACTOR - 1
        if from_cat
        else None
    )

    n_cond = 15000
    # Testing large conditions or dense cells can get quite a lot of halos
    #   so we have an expanded buffer
    buffer_size = 500 * n_cond
    cond_val = options_mass[cond] if from_cat else options_delta[cond]
    cond_array = np.full(n_cond, cond_val)

    sample_dict = cf.sample_halos_from_conditions(
        inputs=inputs,
        redshift=redshift,
        redshift_prev=z_desc,
        cond_array=cond_array,
        buffer_size=buffer_size,
    )

    # set up histogram
    l10min = np.log10(inputs.simulation_options.SAMPLER_MIN_MASS)
    l10max = np.log10(1e14)
    edges = np.linspace(l10min, l10max, num=int(10 * (l10max - l10min))) * np.log(10)
    bin_minima = edges[:-1]
    bin_maxima = edges[1:]
    dlnm = bin_maxima - bin_minima

    # get CMF integrals in the same bins
    binned_cmf = cf.integrate_chmf_interval(
        inputs=inputs,
        lnm_lower=bin_minima,
        lnm_upper=bin_maxima,
        redshift=redshift,
        cond_values=np.array([cond_val]),
        redshift_prev=z_desc,
    ).squeeze()

    hist, _ = np.histogram(np.log(sample_dict["halo_masses"]), edges)

    mass = cond_val if from_cat else cf.get_condition_mass(inputs, "cell")
    mass_dens = (
        inputs.cosmo_params.cosmo.Om0
        * inputs.cosmo_params.cosmo.critical_density(0).to("Mpc-3 M_sun").value
    )
    volume_total_m = mass * n_cond / mass_dens
    one_in_box = 1 / volume_total_m / dlnm

    mf_out = hist / volume_total_m / dlnm
    binned_cmf = binned_cmf * n_cond / volume_total_m / dlnm

    if plt == mpl.pyplot:
        plot_sampler_comparison(
            np.exp(edges),
            sample_dict["expected_progenitors"],
            sample_dict["expected_progenitor_mass"],
            sample_dict["n_progenitors"],
            sample_dict["progenitor_mass"],
            binned_cmf,
            mf_out,
            one_in_box,
            f"mass = {cond_val:.2e}" if from_cat else f"delta = {cond_val:.2e}",
            plt,
        )

    # test the number of progenitors per condition
    np.testing.assert_allclose(
        sample_dict["n_progenitors"].mean(),
        sample_dict["expected_progenitors"][0],
        atol=1,
        rtol=RELATIVE_TOLERANCE,
    )
    # test the total mass of progenitors per condition
    np.testing.assert_allclose(
        sample_dict["progenitor_mass"].mean(),
        sample_dict["expected_progenitor_mass"][0],
        atol=inputs.simulation_options.SAMPLER_MIN_MASS,
        rtol=RELATIVE_TOLERANCE,
    )

    print_failure_stats(
        mf_out,
        binned_cmf,
        [edges[:-1]],
        one_in_box.min(),
        5e-1,
        "binned_cmf",
    )

    np.testing.assert_allclose(
        mf_out,
        binned_cmf,
        atol=2 * one_in_box.min(),  # 2 halo tolerance
        rtol=5e-1,  # 50% tolerance for stochasticity in low-n bins
    )


# NOTE: this test is pretty circular. The only way I think I can test the scaling relations are to
#   calculate them in the backend and re-write them in the test for a few masses. This means that
#   changes to any scaling relation model will result in a test fail
# TODO add minihalo tests, upper turnovers. All 12 properties
def test_halo_prop_sampling(default_input_struct_ts, plt):
    # specify parameters to use for this test
    redshift = 10.0

    # setup the halo masses to test
    halo_mass_vals = np.array([1e5, 1e6, 1e7, 1e8, 1e9, 1e10, 1e11, 1e12])
    halo_rng = np.array([-3, -2, -1, 0, 1, 2, 3])
    halo_masses, halo_rng_in = np.meshgrid(halo_mass_vals, halo_rng, indexing="ij")

    inputs = default_input_struct_ts.evolve_input_structs(
        USE_UPPER_STELLAR_TURNOVER=False,
        M_TURN=5.0,
        F_STAR10=-1,
        ALPHA_STAR=0.0,
        t_STAR=0.1,
        L_X=40.0,
    )
    out_dict = cf.convert_halo_properties(
        redshift=redshift,
        inputs=inputs,
        halo_masses=halo_masses,
        star_rng=halo_rng_in,  # testing the diagonal of the RNG
        sfr_rng=halo_rng_in,
        xray_rng=halo_rng_in,
    )

    halo_mass_out = out_dict["halo_mass"]
    halo_stars_out = out_dict["halo_stars"]
    halo_sfr_out = out_dict["halo_sfr"]
    halo_xray_out = out_dict["halo_xray"]

    # assuming same value for all halos
    ap_c = inputs.astro_params.cdict

    exp_SHMR = (
        ap_c["F_STAR10"]
        * ((halo_masses / 1e10) ** ap_c["ALPHA_STAR"])
        * np.exp(
            -(ap_c["M_TURN"] / halo_masses)
            + (halo_rng_in * ap_c["SIGMA_STAR"])
            - (ap_c["SIGMA_STAR"] ** 2 / 2)
        )
    )
    exp_SHMR = (
        np.minimum(exp_SHMR, 1) * inputs.cosmo_params.OMb / inputs.cosmo_params.OMm
    )
    sim_SHMR = halo_stars_out / halo_mass_out

    sigma_SSFR = ap_c["SIGMA_SFR_LIM"] + ap_c["SIGMA_SFR_INDEX"] * (
        np.log10(exp_SHMR * halo_masses / 1e10)
    )
    sigma_SSFR = np.maximum(sigma_SSFR, ap_c["SIGMA_SFR_LIM"])
    exp_SSFR = (
        inputs.cosmo_params.cosmo.H(redshift).to("s-1").value
        / (ap_c["t_STAR"])
        * np.exp(halo_rng_in * sigma_SSFR - sigma_SSFR**2 / 2)
    )
    sim_SSFR = halo_sfr_out / halo_stars_out

    exp_LX = (
        ap_c["L_X"]
        * np.exp(halo_rng_in * ap_c["SIGMA_LX"] - ap_c["SIGMA_LX"] ** 2 / 2)
        * 1e-38
    )
    sim_LX = halo_xray_out / (halo_sfr_out * 31556925.9747)

    if plt == mpl.pyplot:
        plot_scatter_comparison(
            [exp_SHMR, exp_SSFR, exp_LX],
            [sim_SHMR, sim_SSFR, sim_LX],
            [halo_masses, halo_masses, halo_masses],
            ["SHMR", "SSFR", "LX/SFR"],
            log_inp=True,
            plt=plt,
        )

    print_failure_stats(
        sim_SHMR, exp_SHMR, [halo_mass_vals, halo_rng], 0.0, 1e-4, "SHMR"
    )
    print_failure_stats(
        sim_SSFR,
        exp_SSFR,
        [halo_mass_vals, halo_rng],
        0.0,
        3e-3,  # TODO: fix the difference in t_h between the front and backend
        "SSFR",
    )
    print_failure_stats(sim_LX, exp_LX, [halo_mass_vals, halo_rng], 0.0, 1e-4, "LX")

    np.testing.assert_allclose(exp_SHMR, sim_SHMR, rtol=1e-4)
    np.testing.assert_allclose(exp_SSFR, sim_SSFR, rtol=3e-3)
    np.testing.assert_allclose(exp_LX, sim_LX, rtol=1e-4)


def make_dummy_box(inputs, curr_redshift, redshift_type, output_cls):
    """Make a dummy box of the given output class type.

    All output fields are filled with zeros, and the box is marked
    as having been computed at the given redshift.
    """
    if "curr" in redshift_type:
        out_z = curr_redshift
    elif "prev" in redshift_type:
        out_z = [z for z in inputs.node_redshifts if z > curr_redshift][-1]
    elif "desc" in redshift_type:
        out_z = next([z for z in inputs.node_redshifts if z < curr_redshift])
    else:
        raise ValueError(f"redshift_type {redshift_type} not recognized.")

    box = output_cls.new(inputs, redshift=out_z)
    box._init_arrays()

    for name, array in box.arrays.items():
        setattr(
            box,
            name,
            array.with_value(np.zeros(array.shape, dtype=array.dtype)).computed(),
        )

    return box


def test_perturb_halos(default_input_struct_ts):
    # inputs which get all the firlds
    inputs_test = default_input_struct_ts.evolve_input_structs(
        SOURCE_MODEL="CHMF-SAMPLER",
        SAMPLER_MIN_MASS=5e9,
        PERTURB_ON_HIGH_RES=True,
        INHOMO_RECO=True,
        USE_MINI_HALOS=True,
        USE_RELATIVE_VELOCITIES=True,
        POWER_SPECTRUM="CLASS",
    )
    ics = compute_initial_conditions(
        inputs=inputs_test,
    )
    halofield = determine_halo_catalog(
        redshift=10.0, initial_conditions=ics, inputs=inputs_test
    )

    prev_ts_box = make_dummy_box(
        inputs=inputs_test,
        curr_redshift=10.0,
        redshift_type="prev",
        output_cls=TsBox,
    )
    prev_ion_box = make_dummy_box(
        inputs=inputs_test,
        curr_redshift=10.0,
        redshift_type="prev",
        output_cls=IonizedBox,
    )

    pt_halos = perturb_halo_catalog(
        initial_conditions=ics,
        halo_catalog=halofield,
        previous_spin_temp=prev_ts_box,
        previous_ionize_box=prev_ion_box,
    )

    prop_dict = cf.convert_halo_properties(
        redshift=10.0,
        inputs=inputs_test,
        halo_masses=halofield.get("halo_masses"),
        star_rng=halofield.get("star_rng"),
        sfr_rng=halofield.get("sfr_rng"),
        xray_rng=halofield.get("xray_rng"),
        J_21_LW_grid=prev_ts_box.get("J_21_LW"),
        vcb_grid=ics.get("lowres_vcb"),
        z_re_grid=prev_ion_box.get("z_reion"),
        Gamma12_grid=prev_ion_box.get("ionisation_rate_G12"),
    )

    assert halofield.n_halos == pt_halos.n_halos
    np.testing.assert_allclose(
        pt_halos.get("halo_masses"),
        prop_dict["halo_mass"][: pt_halos.n_halos],
        rtol=1e-5,
    )
    np.testing.assert_allclose(
        pt_halos.get("stellar_masses"),
        prop_dict["halo_stars"][: pt_halos.n_halos],
        rtol=1e-5,
    )
    np.testing.assert_allclose(
        pt_halos.get("sfr"), prop_dict["halo_sfr"][: pt_halos.n_halos], rtol=1e-5
    )
    np.testing.assert_allclose(
        pt_halos.get("ion_emissivity"),
        prop_dict["n_ion"][: pt_halos.n_halos],
        rtol=1e-5,
    )
    np.testing.assert_allclose(
        pt_halos.get("xray_emissivity"),
        prop_dict["halo_xray"][: pt_halos.n_halos],
        rtol=1e-5,
    )
    np.testing.assert_allclose(
        pt_halos.get("fesc_sfr"), prop_dict["halo_wsfr"][: pt_halos.n_halos], rtol=1e-5
    )
    np.testing.assert_allclose(
        pt_halos.get("stellar_mini"),
        prop_dict["halo_stars_mini"][: pt_halos.n_halos],
        rtol=1e-5,
    )
    np.testing.assert_allclose(
        pt_halos.get("sfr_mini"),
        prop_dict["halo_sfr_mini"][: pt_halos.n_halos],
        rtol=1e-5,
    )


# very basic scatter comparison
def plot_scatter_comparison(
    truths, tests, inputs, names, log_vals=True, log_inp=False, plt=None
):
    nplots = len(truths)
    _fig, axs = plt.subplots(nrows=2, ncols=nplots, figsize=(4 * nplots, 6))

    for i, (true, test, inp, name) in enumerate(
        zip(truths, tests, inputs, names, strict=False)
    ):
        axs[0, i].scatter(true, test, s=2, rasterized=True)
        axs[0, i].set_xlabel(f"True {name}")
        axs[0, i].set_ylabel(f"Test {name}")
        axs[0, i].grid()

        axs[1, i].scatter(inp, test / true, s=5, rasterized=True)
        axs[1, i].set_xlabel("input value")
        axs[1, i].set_ylabel(f"Test/True {name}")
        axs[1, i].grid()

        if log_inp:
            axs[1, i].set_xscale("log")
        if log_vals:
            axs[1, i].set_yscale("log")
            axs[0, i].set_xscale("log")
            axs[0, i].set_yscale("log")


def plot_sampler_comparison(
    bin_edges, exp_N, exp_M, N_array, M_array, exp_mf, mf_out, one_halo, title, plt
):
    _fig, axs = plt.subplots(nrows=1, ncols=2, figsize=(12, 5))

    axst = [axs[0].twinx(), axs[1].twiny()]

    # total M axis
    axs[1].set_xlabel("M")
    # total N axis
    axst[1].set_ylabel("N")

    # mass function axis
    axs[0].set_title(title)
    axs[0].set_ylim([one_halo[0] / 2, exp_mf.max()])
    axs[0].set_yscale("log")
    axs[0].set_ylabel("dn/dlnM")
    axs[0].set_xlabel("M")

    # ratio axis
    axst[0].set_ylim([1e-1, 1e1])
    axst[0].set_yscale("log")
    axst[0].set_ylabel("ratio")

    for ax in axs:
        ax.grid()
        ax.set_xscale("log")

    # log-spaced bins
    dlnm = np.log(bin_edges[1:]) - np.log(bin_edges[:-1])
    bin_centres = (bin_edges[:-1] * np.exp(dlnm / 2)).astype("f4")
    edges_n = np.arange(0, max(N_array.max(), 2), 1)
    centres_n = (edges_n[:-1] + edges_n[1:]) / 2

    hist_n, _ = np.histogram(N_array, edges_n)
    p_n = hist_n / N_array.size
    hist_m, _ = np.histogram(M_array, bin_edges)
    p_m = hist_m / M_array.size / dlnm  # p(lnM)

    axst[0].loglog(bin_centres, mf_out / exp_mf, color="r", linewidth=2)
    axst[0].loglog(
        bin_centres, np.ones_like(exp_mf), color="r", linestyle=":", linewidth=1
    )

    axs[0].loglog(bin_centres, mf_out, color="k", linewidth=3, label="Sample")
    axs[0].loglog(
        bin_centres, exp_mf, color="k", linestyle=":", linewidth=1, label="Expected"
    )
    axs[0].loglog(
        bin_centres, one_halo, color="k", linestyle="--", linewidth=2, label="One Halo"
    )
    axs[0].legend()

    axs[1].semilogx(bin_centres, p_m / p_m.max(), color="k", linewidth=2)
    axs[1].axvline(exp_M[0], color="k", linestyle=":", linewidth=2)
    axs[1].axvline(M_array.mean(), color="k", linestyle="-", linewidth=1)
    axs[1].set_xlim(M_array.min() / 10, M_array.max() * 10)

    axst[1].plot(centres_n, p_n / p_n.max(), color="r", linewidth=2)
    axst[1].axvline(exp_N[0], color="r", linestyle=":", linewidth=2)
    axst[1].axvline(N_array.mean(), color="r", linestyle="-", linewidth=1)
    axst[1].set_xlim(N_array.min() - 1, N_array.max() + 1)



"""Classes dealing with the state of arrays."""

from typing import Self

import attrs


class ArrayStateError(ValueError):
    """Errors arising from incorrectly modifying array state."""


@attrs.define(frozen=True)
class ArrayState:
    """Define the memory state of a struct array."""

    initialized = attrs.field(default=False, converter=bool)
    c_memory = attrs.field(default=False, converter=bool)
    computed_in_mem = attrs.field(default=False, converter=bool)
    on_disk = attrs.field(default=False, converter=bool)

    def deinitialize(self) -> Self:
        """Return new state that is not initialized."""
        return attrs.evolve(self, initialized=False, computed_in_mem=False)

    def initialize(self) -> Self:
        """Return new state that is initialized."""
        return attrs.evolve(self, initialized=True)

    def computed(self) -> Self:
        """Return new state indicating the array has been computed."""
        return attrs.evolve(self, computed_in_mem=True, initialized=True)

    def dropped(self) -> Self:
        """Return new state indicating the array has been dropped from memory."""
        return attrs.evolve(self, initialized=False, computed_in_mem=False)

    def written(self) -> Self:
        """Return new state indicating the array has been written to disk."""
        return attrs.evolve(self, on_disk=True)

    def purged_to_disk(self) -> Self:
        """Return new state indicating the array has been written to disk and dropped."""
        return self.written().dropped()

    def loaded_from_disk(self) -> Self:
        """Return new state indicating the array has been loaded from disk into memory."""
        return self.computed().written()

    @property
    def is_computed(self) -> bool:
        """Whether the array is computed anywhere."""
        return self.computed_in_mem or self.on_disk

    @property
    def c_has_active_memory(self) -> bool:
        """Whether C currently has initialized memory for this array."""
        return self.c_memory and self.initialized

    def __str__(self) -> str:
        """Return a string representation of the ArrayState."""
        if self.computed_in_mem:
            return "computed (in mem)"
        elif self.on_disk:
            return "computed (on disk)"
        elif self.initialized:
            return "memory initialized (not computed)"
        else:
            return "uncomputed and uninitialized"



"""
Large-scale tests which test code updates against previously-run "golden" results.

The idea here is that any new updates (except for major versions) should be non-breaking;
firstly, they should not break the API, so that the tests should run without crashing without
being changed.
Secondly, the actual results of running the basic functions should remain the same for
the same input code, except for potential bug-fixes. In these cases, these tests should
pick these changes up. The test data should then be changed to reflect the new gold
standard, and if applicable, a new test should be written that reflects the previous
broken code.
Thirdly, it enforces that new features, where possible, are added in such a way as to
keep the default behaviour constant. That is, the tests here should *not* run the added
feature, and therefore should continue to produce the same test results regardless of
the new feature added. The new feature should be accompanied by its own tests, whether
in this or another test module. If a new feature *must* be included by default, then
it must be implemented in a new major version of the code, at which point the test data
is able to be updated.

Comparison tests here are meant to be as small as possible while attempting to form
a reasonable test: they should be of reduced data such as power spectra or global xHI
measurements, and they should be generated with small simulations.
"""

import logging

import h5py
import matplotlib as mpl
import numpy as np
import pytest

from py21cmfast import Coeval, LightCone, OutputCache
from py21cmfast.wrapper.inputs import Table1D

from . import produce_integration_test_data as prd

logger = logging.getLogger("py21cmfast")
logger.setLevel(logging.INFO)

options = list(prd.OPTIONS_TESTRUNS.keys())
options_pt = list(prd.OPTIONS_PT.keys())


@pytest.mark.parametrize("name", options)
def test_power_spectra_coeval(name, module_direc, plt):
    redshift, kwargs = prd.OPTIONS_TESTRUNS[name]
    print(f"Options used for the test {name} at z={redshift}: ", kwargs)

    # First get pre-made data
    true_powers = {}
    with h5py.File(prd.get_filename("power_spectra", name), "r") as fl:
        true_k = fl["coeval"]["k"][...]
        for key, value in fl["coeval"].items():
            if key.startswith("power_"):
                fieldname = "_".join(key.split("_")[1:])
                true_powers[fieldname] = value[...]
                print(f"{key} --> {fieldname} loaded")

    # Now compute the Coeval object
    test_k, test_powers, cv = prd.produce_coeval_power_spectra(
        redshift, cache=OutputCache(module_direc), **kwargs
    )

    assert isinstance(cv, Coeval)
    assert np.all(np.isfinite(cv.brightness_temp))

    assert np.allclose(true_k, test_k)

    any_failed = False
    # We don't assert that all the fields are identical, but print the differences
    for key in prd.COEVAL_FIELDS:
        if key not in true_powers:
            continue
        any_failed |= prd.print_failure_stats(
            test_powers[key],
            true_powers[key],
            [test_k],
            abs_tol=0,
            rel_tol=1e-4,
            name=key,
        )

    if plt == mpl.pyplot and any_failed:
        make_coeval_comparison_plot(true_k, test_k, true_powers, test_powers, plt)


@pytest.mark.parametrize("name", options)
def test_power_spectra_lightcone(name, module_direc, plt, benchmark):
    redshift, kwargs = prd.OPTIONS_TESTRUNS[name]
    print(f"Options used for the test {name} at z={redshift}: ", kwargs)

    # First get pre-made data
    with h5py.File(prd.get_filename("power_spectra", name), "r") as fl:
        true_powers = {}
        true_global = {}
        true_k = fl["lightcone"]["k"][...]
        for key in fl["lightcone"]:
            fieldname = "_".join(key.split("_")[1:])
            if key.startswith("power_"):
                true_powers[fieldname] = fl["lightcone"][key][...]
            elif key.startswith("global_"):
                true_global[fieldname] = fl["lightcone"][key][...]

    # Now compute the lightcone
    test_k, test_powers, lc = benchmark.pedantic(
        prd.produce_lc_power_spectra,
        kwargs=dict(redshift=redshift, cache=OutputCache(module_direc), **kwargs),
        iterations=1,  # these tests can be slow
        rounds=1,
    )

    if lc.inputs.matter_options.POWER_SPECTRUM == "CLASS":
        with h5py.File(prd.get_filename("power_spectra", name), "r") as fl:
            x_values = fl["cosmo_tables"]["transfer_density"]["x_values"]
            y_values = fl["cosmo_tables"]["transfer_density"]["y_values"]
            transfer_density = Table1D(
                size=x_values.size, x_values=x_values, y_values=y_values
            )
        assert transfer_density.x_values.size == transfer_density.y_values.size
        np.testing.assert_allclose(
            transfer_density.x_values,
            lc.inputs.cosmo_tables.transfer_density.x_values,
            atol=0,
            rtol=1e-3,
        )
        np.testing.assert_allclose(
            transfer_density.y_values,
            lc.inputs.cosmo_tables.transfer_density.y_values,
            atol=0,
            rtol=5e-3,
        )

    assert isinstance(lc, LightCone)
    assert np.all(np.isfinite(lc.lightcones["brightness_temp"]))

    test_global = {k: lc.global_quantities[k] for k in true_global}
    assert np.allclose(true_k, test_k)

    # We don't assert that all the fields are identical, but print the differences
    any_failed = False
    for key in prd.LIGHTCONE_FIELDS:
        if key not in true_powers:
            continue
        any_failed |= prd.print_failure_stats(
            test_powers[key],
            true_powers[key],
            [test_k],
            abs_tol=0,
            rel_tol=1e-4,
            name=key,
        )

    any_failed = True  # TODO:remove this testing line
    if plt == mpl.pyplot and any_failed:
        make_lightcone_comparison_plot(
            true_k,
            test_k,
            lc.inputs.node_redshifts,
            true_powers,
            true_global,
            test_powers,
            test_global,
            plt,
        )

    # For globals, we should assert that they are close
    for key, value in true_global.items():
        np.testing.assert_allclose(value, lc.global_quantities[key], atol=0, rtol=1e-3)


plot_ylab = {
    "density": r"$\delta$",
    "lowres_density": r"$\delta_{IC}$",
    "lowres_vx_2LPT": r"$v_x$ IC 2LPT",
    "lowres_vx": r"$v_x$ IC",
    "velocity_z": r"$v_z$",
    "spin_temperature": r"$T_S$",
    "xray_ionised_fraction": r"$x_{e}$",
    "J_21_LW": r"$J_{21,LW}$",
    "kinetic_temp_neutral": r"$T_{K,HI}$",
    "ionisation_rate_G12": r"$\Gamma_{12}$",
    "cumulative_recombinations": r"$N_{rec}$",
    "neutral_fraction": r"$x_{HI}$",
    "z_reion": r"$z_{re}$",
    "brightness_temp": r"$dT_b$",
}


def make_lightcone_comparison_plot(
    true_k, k, z, true_powers, true_global, test_powers, test_global, plt
):
    n = len(true_global) + len(true_powers)
    _fig, ax = plt.subplots(
        2,
        n,
        figsize=(4 * n, 6),
        constrained_layout=True,
        sharex="col",
        gridspec_kw={"hspace": 0.1, "wspace": 0.1},
    )

    for i, (key, val) in enumerate(test_powers.items()):
        make_comparison_plot(
            true_k,
            k,
            true_powers[key],
            val,
            ax[:, i],
            xlab="k",
            ylab=f"{plot_ylab[key]} Power",
            make_lower_ylab=i == 0,
        )

    for j, (key, val) in enumerate(test_global.items(), start=i + 1):
        make_comparison_plot(
            z,
            z,
            true_global[key],
            val,
            ax[:, j],
            xlab="z",
            ylab=f"{plot_ylab[key]}",
            logx=False,
            logy=False,
            make_lower_ylab=False,
        )


def make_coeval_comparison_plot(true_k, k, true_powers, test_powers, plt):
    _fig, ax = plt.subplots(
        2,
        len(true_powers),
        figsize=(4 * len(true_powers), 6),
        sharex=True,
        constrained_layout=True,
    )

    for i, (key, val) in enumerate(test_powers.items()):
        make_comparison_plot(
            true_k,
            k,
            true_powers[key],
            val,
            ax[:, i],
            xlab="k",
            ylab=f"{plot_ylab[key]} Power",
            make_lower_ylab=i == 0,
        )


def make_comparison_plot(
    xtrue,
    x,
    true,
    test,
    ax,
    logx=True,
    logy=True,
    xlab=None,
    ylab=None,
    make_lower_ylab=True,
):
    ax[0].plot(xtrue, true, label="True")
    ax[0].plot(x, test, label="Test")
    if logx:
        ax[0].set_xscale("log")
    if logy:
        ax[0].set_yscale("log")
    if xlab:
        ax[1].set_xlabel(xlab)
    if ylab:
        ax[0].set_title(ylab)

    ax[0].legend()

    ax[1].plot(x, (test - true) / true)

    if make_lower_ylab:
        ax[1].set_ylabel("Fractional Difference")


@pytest.mark.parametrize("name", options_pt)
def test_perturb_field_data(name):
    redshift, kwargs = prd.OPTIONS_PT[name]
    print("Options used for the test: ", kwargs)

    # First get pre-made data
    with h5py.File(prd.get_filename("perturb_field_data", name), "r") as f:
        power_dens = f["power_dens"][...]
        power_vel = f["power_vel"][...]
        pdf_dens = f["pdf_dens"][...]
        pdf_vel = f["pdf_vel"][...]

    (
        _k_dens,
        p_dens,
        _k_vel,
        p_vel,
        _x_dens,
        y_dens,
        _x_vel,
        y_vel,
        _ic,
    ) = prd.produce_perturb_field_data(redshift, **kwargs)

    np.testing.assert_allclose(p_dens, power_dens, atol=5e-3, rtol=1e-3)
    np.testing.assert_allclose(p_vel, power_vel, atol=5e-3, rtol=1e-3)
    np.testing.assert_allclose(y_dens, pdf_dens, atol=5e-3, rtol=1e-3)
    np.testing.assert_allclose(y_vel, pdf_vel, atol=5e-3, rtol=1e-3)
